/**
 * Insertion sort on a doubly-linked list.
 * In main(), call either test() or measure().
 * measure(), if the corresponding code in List.sort() is not commented out,
 * will produce on standard output a table that can be used to visualize the complexity.
 * See measurements.txt for an example table, and measurements.pdf for the resulting plot.
 * See measurements.R for the R script that produces the plot.
 * 
 * @author Matthias.Hauswirth@usi.ch
 */
package ch.usi.inf.sape.algoprof.example.sort.insertion;

import java.util.Random;

import ch.usi.inf.sape.algoprof.statistics.Algorithm;
import ch.usi.inf.sape.algoprof.statistics.AlgorithmCounterVisitor;
import ch.usi.inf.sape.algoprof.statistics.AlgorithmRepetitionTreeVisitor;
import ch.usi.inf.sape.algoprof.statistics.AlgorithmVisitor;
import ch.usi.inf.sape.algoprof.statistics.Configuration;


public class Main implements Algorithm {

	private AlgorithmCounterVisitor constructVisitor = null;
	private AlgorithmCounterVisitor constructReversedVisitor = null;
	private AlgorithmCounterVisitor constructRandomVisitor = null;
	private AlgorithmCounterVisitor sortVisitor = null;
	private AlgorithmCounterVisitor measureVisitor = null;
	private AlgorithmCounterVisitor printVisitor = null;
	
	private int counter = 0;
	
	public static void main(final String[] args) {
		//test();
		new Main().measure();
	}
	
	private void test() {
		final List list = new List();
		final int size = 10;
		//constructReversed(list, size); // worst case input
		//construct(list, size); // best case input
		constructRandom(list, size); // expected case input (assuming element values are expected to be uniformly distributed random numbers)
		print(list);
		sort(list);
		print(list);
	}
	
	private void measure() {
		
		boolean once = true;
		
		AlgorithmVisitor av = null;
		if (Configuration.ENABLE_MANUAL_INSTRUMENTATION) {
			//constructVisitor = new AlgorithmCounterVisitor("construct", "/src/ch/usi/inf/sape/algoprof/example/sort/insertion/", "List Node construction");
			//constructReversedVisitor = new AlgorithmCounterVisitor("constructReversed", "/src/ch/usi/inf/sape/algoprof/example/sort/insertion/", "List Node construction");
			constructRandomVisitor = new AlgorithmCounterVisitor("constructRandom", "/src/ch/usi/inf/sape/algoprof/example/sort/insertion/", "List Node construction");
			sortVisitor = new AlgorithmCounterVisitor("sort", "/src/ch/usi/inf/sape/algoprof/example/sort/insertion/", "List Node modification");
			measureVisitor = new AlgorithmCounterVisitor("measure", "/src/ch/usi/inf/sape/algoprof/example/sort/insertion/", "int size traversal");
			//printVisitor = new AlgorithmCounterVisitor("print", "/src/ch/usi/inf/sape/algoprof/example/sort/insertion/", "List Node traversal");
			
			av = new AlgorithmRepetitionTreeVisitor();
			accept(av);
		}
		
		// sizes 0 to 1000, 10 runs for each size, each run will sort with the given number of random elements
		// this should print out a table that can be plotted to check the average complexity of the sort method
		System.out.println("Size\tComparisons");
		//int n[] = new int [] {2,5,10,15,20,25,40,50};//,80,100,150,200};
		for (int size=0; size<50; size=size+10) {
		//for (int size : n) {
			for (int i=0; i<3; i++) {
				final List list = new List();
				if (once && Configuration.ENABLE_MANUAL_INSTRUMENTATION) {
					once = false;
					list.accept(av);
				}
				//constructReversed(list, size);
				//if (Configuration.ENABLE_MANUAL_INSTRUMENTATION) {
				//	list.setData(constructReversedVisitor, counter);
				//}
				//construct(list, size);
				//if (Configuration.ENABLE_MANUAL_INSTRUMENTATION) {
				//	list.setData(constructVisitor, counter); 
				//}
				constructRandom(list, size);
				if (Configuration.ENABLE_MANUAL_INSTRUMENTATION) {
					list.setData(constructRandomVisitor, counter);
				}
				System.out.println(size);
				sort(list);
				if (Configuration.ENABLE_MANUAL_INSTRUMENTATION) {
					list.accept(sortVisitor);
				}
			}
		
		}
		
		if (Configuration.ENABLE_MANUAL_INSTRUMENTATION) {
			//accept(constructVisitor);
			//accept(constructReversedVisitor);
			accept(constructRandomVisitor);
			accept(sortVisitor);
			accept(measureVisitor);
			//accept(printVisitor);
		}
	}
	
	private void construct(List list, int size) {
		counter = 0;
		for (int i=0; i<size; i++) {
			counter++;
			list.append(i);
		}
	}
	
	private void constructReversed(List list, int size) {
		counter = 0;
		for (int i=0; i<size; i++) {
			counter++;
			list.append(9-i);
		}
	}
	
	private void constructRandom(List list, int size) {
		counter = 0;
		Random r = new Random();
		for (int i=0; i<size; i++) {
			counter++;
			list.append(r.nextInt(size));
		}
	}
	
	private void print(List list) {
		list.print();
	}

	private static void sort(List list) {
		list.sort();
	}

	public void accept(AlgorithmVisitor visitor) {
		visitor.visit(this);
		
	}
	
}
