package ch.unisi.inf.sp.cfg;

import static org.objectweb.asm.Opcodes.ATHROW;

import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.FieldInsnNode;
import org.objectweb.asm.tree.IincInsnNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.IntInsnNode;
import org.objectweb.asm.tree.JumpInsnNode;
import org.objectweb.asm.tree.LabelNode;
import org.objectweb.asm.tree.LdcInsnNode;
import org.objectweb.asm.tree.LookupSwitchInsnNode;
import org.objectweb.asm.tree.MethodInsnNode;
import org.objectweb.asm.tree.MethodNode;
import org.objectweb.asm.tree.MultiANewArrayInsnNode;
import org.objectweb.asm.tree.TableSwitchInsnNode;
import org.objectweb.asm.tree.TryCatchBlockNode;
import org.objectweb.asm.tree.TypeInsnNode;
import org.objectweb.asm.tree.VarInsnNode;
import org.objectweb.asm.util.AbstractVisitor;

/**
 * This class extracts a control flow graph in .dot format from the byte code of
 * a Java method.
 * 
 * @author Pavel Raykov
 */
public final class ControlFlowGraphExtractor {
	
	public BufferedWriter bf;
	
	@SuppressWarnings("unchecked")
	public void build(ClassNode clazz, BufferedWriter bw) throws IOException {

		bf = bw;
		bf.write("digraph G {\n");
		//System.out.println("build started " + clazz.name);
		final List<MethodNode> methods = clazz.methods;
		int i = 0;
		for (MethodNode method : methods) {
			disassembleMethod(method, i++);
		}
		bf.write("}");
		bf.flush();
		bf.close();
	}

	private void disassembleMethod(final MethodNode method, int order) {
		Exceptions exceptions = new Exceptions();
		
		for (final Object o : method.tryCatchBlocks) {
			final TryCatchBlockNode node = (TryCatchBlockNode)o;
			InsnList instrs = method.instructions;
			exceptions.add(new ExceptionData(
					instrs.indexOf(node.start),
					instrs.indexOf(node.end),
					instrs.indexOf(node.handler),
					node.type));
		}
		
		final InsnList instructions = method.instructions;
		final List<InstructionData> instData = getInstructions(instructions);
		final InstType[] jumps = makeJumps(instData, exceptions);
		List<Block> blocks = makeBlocks(instData, jumps);
		outputGraph(instructions.size(), blocks, method.name + method.desc, order);
	}

	//ALL THE SUFFIX STAFF WAS ADDED BECAUSE IT LOOKS THAT DOT LANGUAGE DOES NOT SUPPORT NAMESPACES =(
	private void outputGraph(final int n, List<Block> blocks, String graphName, int order) {
		StringBuffer blocksBuff = new StringBuffer();
		StringBuffer edgesBuff = new StringBuffer();
		String suffix = "xxx" + order;
		//Adds start and end
		edgesBuff.append("b" + n + suffix + "->" + "b0" + suffix + "\n");
		blocksBuff.append("b" + n + suffix +"[shape=record,  label = \"S\"]\n" );
		blocksBuff.append("b" + (n+1) + suffix + "[shape=record,  label = \"E\"]\n" );
		
		for (int i = 0; i < blocks.size(); i++) {
			blocksBuff.append(blocks.get(i).getInstructions(suffix));
			edgesBuff.append(blocks.get(i).getEdgesAsString(suffix));
		}
		try {
			//System.out.println("subgraph cluster" + order + " { label = \"" + graphName + "\"\n" + blocksBuff.append(edgesBuff) + "}\n");
			bf.write("subgraph cluster" + order + " { label = \"" + graphName + "\"\n" + blocksBuff.append(edgesBuff) + "}\n");
		} catch (IOException e) {
			System.err.println("IOException when building a subgraph");
			e.printStackTrace();
		}
	}

	private List<InstructionData> getInstructions(InsnList instructions) {
		final int n = instructions.size();
		final List<InstructionData> instData = new ArrayList<InstructionData>(n);
		for (int i = 0; i < n; i++) {
			final AbstractInsnNode instruction = instructions.get(i);
			InstructionData data = disassembleInstruction(instruction, i, instructions);
			if (data == null) {
				throw new IllegalStateException("Unknown instruction - " + instruction.getOpcode());
			}
			instData.add(data);
		}
		return instData;
	}
	
	private List<Block> makeBlocks(final List<InstructionData> instData, final InstType[] jumps) {
		// Make blocks
		int n = instData.size();
		List<Block> blocks = new ArrayList<Block>(n);

		Block current = new Block();
		for (int i = 0; i < n; i++) {
			switch (jumps[i]) {
				case NORMAL:
					current.addInstr(instData.get(i));
					break;
				case IN:
					if (!current.isEmpty()) {
						blocks.add(current);
						if (current.getEdges().isEmpty()) {
							current.getEdges().add(new NormalEdge(i, ""));
						}
					}
					current = new Block();
					current.addInstr(instData.get(i));
					break;
				case OUT:
					current.addInstr(instData.get(i));
					current.setEdges(instData.get(i).edges);
					blocks.add(current);
					current = new Block();
					break;
				case RETURN:
					current.addInstr(instData.get(i));
					blocks.add(current);
					current = new Block();
					break;
			}
		}
		if (!current.isEmpty()) {
			blocks.add(current);
		}
		return blocks;
	}

	private InstType[] makeJumps(List<InstructionData> instData, Exceptions exceptions) {
		// Make jumps
		int n = instData.size();
		final InstType[] jumps = new InstType[n];
		for (int i = 0; i < n; i++) {
			jumps[i] = InstType.NORMAL;
		}

		for (int i = 0; i < n; i++) {
			InstructionData data = instData.get(i);
			if (data.isReturn()) {
				jumps[i] = InstType.OUT;
				data.edges.add(new NormalEdge(n + 1,""));
			} else if (data.edges.size() > 0) {
				jumps[i] = InstType.OUT;
				for (Edge edge : data.edges) {
					jumps[edge.t] = InstType.IN;
				}
			}
			
			if (data.isPEI()) {
				jumps[i] = (jumps[i] == InstType.RETURN) ? InstType.RETURN : InstType.OUT;
				Pair<List<Integer>, Boolean> pair = exceptions.getPossibleJumps(data.i);
				List<Integer> exJumps = pair.t;
				for (int j: exJumps) {
					jumps[j] = InstType.IN;
					data.edges.add(new ExceptionEdge(j, "ex"));
				}
				if (!pair.v) {
					data.edges.add(new ExceptionEdge(n + 1, "ex"));
				}
				if ((data.opcode != ATHROW) && (!data.isReturn()) && (i+1<n)) {
					jumps[i+1] = InstType.IN;
					data.edges.add(new NormalEdge(i+1, ""));
				}
			}
		}
		return jumps;
	}

	/**
	 * Hint: Check out org.objectweb.asm.MethodVisitor to determine which
	 * instructions (opcodes) have which instruction types (subclasses of
	 * AbstractInsnNode).
	 * 
	 * E.g. the comment in org.objectweb.asm.MethodVisitor.visitIntInsn(int
	 * opcode, int operand) shows the list of all opcodes that are represented
	 * as instructions of type IntInsnNode. That list e.g. includes the BIPUSH
	 * opcode.
	 */
	@SuppressWarnings("unchecked")
	public InstructionData disassembleInstruction(
			final AbstractInsnNode instruction, 
			final int i,
			final InsnList instructions) {
		final int opcode = instruction.getOpcode();
		final String mnemonic = opcode == -1 ? ""
				: AbstractVisitor.OPCODES[instruction.getOpcode()] + " ";
		// There are different subclasses of AbstractInsnNode.
		// AbstractInsnNode.getType() represents the subclass as an int.
		// Note:
		// to check the subclass of an instruction node, we can either use:
		// if (instruction.getType()==AbstractInsnNode.LABEL)
		// or we can use:
		// if (instruction instanceof LabelNode)
		// They give the same result, but the first one can be used in a switch
		// statement.
		switch (instruction.getType()) {
		case AbstractInsnNode.LABEL:
			// pseudo-instruction (branch or exception target)
			return new InstructionData(i, opcode, i);
		case AbstractInsnNode.FRAME:
			// pseudo-instruction (stack frame map)
			return new InstructionData(i, opcode, i);
		case AbstractInsnNode.LINE:
			// pseudo-instruction (line number information)
			return new InstructionData(i, opcode, i);
		case AbstractInsnNode.INSN:
			// Opcodes: NOP, ACONST_NULL, ICONST_M1, ICONST_0, ICONST_1,
			// ICONST_2,
			// ICONST_3, ICONST_4, ICONST_5, LCONST_0, LCONST_1, FCONST_0,
			// FCONST_1, FCONST_2, DCONST_0, DCONST_1, IALOAD, LALOAD, FALOAD,
			// DALOAD, AALOAD, BALOAD, CALOAD, SALOAD, IASTORE, LASTORE,
			// FASTORE,
			// DASTORE, AASTORE, BASTORE, CASTORE, SASTORE, POP, POP2, DUP,
			// DUP_X1, DUP_X2, DUP2, DUP2_X1, DUP2_X2, SWAP, IADD, LADD, FADD,
			// DADD, ISUB, LSUB, FSUB, DSUB, IMUL, LMUL, FMUL, DMUL, IDIV, LDIV,
			// FDIV, DDIV, IREM, LREM, FREM, DREM, INEG, LNEG, FNEG, DNEG, ISHL,
			// LSHL, ISHR, LSHR, IUSHR, LUSHR, IAND, LAND, IOR, LOR, IXOR, LXOR,
			// I2L, I2F, I2D, L2I, L2F, L2D, F2I, F2L, F2D, D2I, D2L, D2F, I2B,
			// I2C, I2S, LCMP, FCMPL, FCMPG, DCMPL, DCMPG, IRETURN, LRETURN,
			// FRETURN, DRETURN, ARETURN, RETURN, ARRAYLENGTH, ATHROW,
			// MONITORENTER, or MONITOREXIT.
			// zero operands, nothing to print
			return new InstructionData(mnemonic, i, opcode, i);
		case AbstractInsnNode.INT_INSN:
			// Opcodes: NEWARRAY, BIPUSH, SIPUSH.
			if (instruction.getOpcode() == Opcodes.NEWARRAY) {
				// NEWARRAY
				return new InstructionData(mnemonic + AbstractVisitor.TYPES[((IntInsnNode) instruction).operand], i, opcode, i);
			} else {
				// BIPUSH or SIPUSH
				return new InstructionData(mnemonic + ((IntInsnNode) instruction).operand, i, opcode, i);
			}
		case AbstractInsnNode.JUMP_INSN:
			// Opcodes: IFEQ, IFNE, IFLT, IFGE, IFGT, IFLE, IF_ICMPEQ,
			// IF_ICMPNE, IF_ICMPLT, IF_ICMPGE, IF_ICMPGT, IF_ICMPLE, IF_ACMPEQ,
			// IF_ACMPNE, GOTO, JSR, IFNULL or IFNONNULL.
		{
			final LabelNode targetInstruction = ((JumpInsnNode) instruction).label;
			final int targetId = instructions.indexOf(targetInstruction);
			
			List<Edge> edges = new ArrayList<Edge>();
			
			if (!mnemonic.equals("GOTO ")) {
				edges.add(new NormalEdge(i+1, "F"));
				edges.add(new NormalEdge(targetId, "T"));
			} else {
				edges.add(new NormalEdge(targetId, ""));
			}
			return new InstructionData(edges, mnemonic + targetId, i, opcode, i);
		}
		case AbstractInsnNode.LDC_INSN:
			// Opcodes: LDC.
			return new InstructionData(mnemonic + ((LdcInsnNode) instruction).cst, i, opcode, i);
		case AbstractInsnNode.IINC_INSN:
			// Opcodes: IINC.
			return new InstructionData(mnemonic + ((IincInsnNode) instruction).var + " " 
					+ ((IincInsnNode) instruction).incr, i, opcode, i);
		case AbstractInsnNode.TYPE_INSN:
			// Opcodes: NEW, ANEWARRAY, CHECKCAST or INSTANCEOF.
			return new InstructionData(mnemonic + ((TypeInsnNode) instruction).desc, i, opcode, i);
		case AbstractInsnNode.VAR_INSN:
			// Opcodes: ILOAD, LLOAD, FLOAD, DLOAD, ALOAD, ISTORE,
			// LSTORE, FSTORE, DSTORE, ASTORE or RET.
			return new InstructionData(mnemonic + ((VarInsnNode) instruction).var, i, opcode, i);
		case AbstractInsnNode.FIELD_INSN:
			// Opcodes: GETSTATIC, PUTSTATIC, GETFIELD or PUTFIELD.
			return new InstructionData(mnemonic + ((FieldInsnNode) instruction).owner + "." 
					+ ((FieldInsnNode) instruction).name + " " + ((FieldInsnNode) instruction).desc, i, opcode, i);
		case AbstractInsnNode.METHOD_INSN:
			// Opcodes: INVOKEVIRTUAL, INVOKESPECIAL, INVOKESTATIC,
			// INVOKEINTERFACE or INVOKEDYNAMIC.
			return new InstructionData(mnemonic + ((MethodInsnNode) instruction).owner + "."
					+ ((MethodInsnNode) instruction).name + " " + ((MethodInsnNode) instruction).desc, i, opcode, i);
		case AbstractInsnNode.MULTIANEWARRAY_INSN:
			// Opcodes: MULTIANEWARRAY.
			return new InstructionData(mnemonic + ((MultiANewArrayInsnNode) instruction).desc + " " +
					((MultiANewArrayInsnNode) instruction).dims, i, opcode, i);
		case AbstractInsnNode.LOOKUPSWITCH_INSN:
			// Opcodes: LOOKUPSWITCH.
		{
			List<Edge> edges = new ArrayList<Edge>();
			final List keys = ((LookupSwitchInsnNode) instruction).keys;
			final List labels = ((LookupSwitchInsnNode) instruction).labels;
			for (int t = 0; t < keys.size(); t++) {
				final int key = (Integer) keys.get(t);
				final LabelNode targetInstruction = (LabelNode) labels.get(t);
				final int targetId = instructions.indexOf(targetInstruction);
				edges.add(new NormalEdge(targetId, "K:" + key));
			}
			final LabelNode defaultTargetInstruction = ((LookupSwitchInsnNode) instruction).dflt;
			final int defaultTargetId = instructions
					.indexOf(defaultTargetInstruction);
			edges.add(new NormalEdge(defaultTargetId, "default:"));
			return new InstructionData(edges, mnemonic, i, opcode, i);
		}
		case AbstractInsnNode.TABLESWITCH_INSN:
			// Opcodes: TABLESWITCH.
		{
			List<Edge> edges = new ArrayList<Edge>();
			final int minKey = ((TableSwitchInsnNode) instruction).min;
			final List labels = ((TableSwitchInsnNode) instruction).labels;
			for (int t = 0; t < labels.size(); t++) {
				final int key = minKey + t;
				final LabelNode targetInstruction = (LabelNode) labels.get(t);
				final int targetId = instructions.indexOf(targetInstruction);
				edges.add(new NormalEdge(targetId, "K:" + key));
			}
			final LabelNode defaultTargetInstruction = ((TableSwitchInsnNode) instruction).dflt;
			final int defaultTargetId = instructions
					.indexOf(defaultTargetInstruction);
			edges.add(new NormalEdge(defaultTargetId, "default:"));
			return new InstructionData(edges, mnemonic, i, opcode, i);
		}
		}
		return null;
	}

	public String myMethodToString(MethodNode method) {
		return method.name + method.desc;
	}
}
