package ch.unisi.inf.sp.type;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.tree.ClassNode;

import usi.instrumentation.InlineJSR;


/**
 * Scans a JAR archive containing Java class files, uses ASM to load each class,
 * and for each class invokes Analyzer.analyze() on each registered Analyzer.
 *
 * @author Matthias.Hauswirth@usi.ch
 */
public final class ArchiveScanner {

	private final ArrayList<ClassAnalyzer> analyzers;
	
	
	public ArchiveScanner() {
		analyzers = new ArrayList<ClassAnalyzer>();
	}
	
	public void addAnalyzer(final ClassAnalyzer analyzer) {
		analyzers.add(analyzer);
	}
	
	public void removeAnalyzer(final ClassAnalyzer analyzer) {
		analyzers.remove(analyzer);
	}
	
	private boolean isClass (final String name) {
		File f = new File (name);
		try {
			final ZipFile zipFile = new ZipFile(name);
		} catch (IOException e) {
			return true;
		}
		return false;
	}
	
	public void scan(final String archiveName) throws IOException {
		
		if (archiveName.endsWith(".class") || isClass (archiveName))  {
			//String name = archiveName.substring(archiveName.l, endIndex)
			
				analyzeClass(archiveName);
			
		} else {
			//System.out.println(archiveName);
			File file = new File(archiveName);
			if (file.exists()) {
				final ZipFile zipFile = new ZipFile(archiveName);
				final Enumeration<? extends ZipEntry> entries = zipFile.entries();
				while (entries.hasMoreElements()) {
					final ZipEntry zipEntry = entries.nextElement();
					if (zipEntry.getName().toLowerCase().endsWith(".class")) {
						//System.out.println(zipEntry.getName());
						analyzeClass(zipFile, zipEntry);
					}
				}
				
			}
			
		}
	}
	
	private HashMap<String, ClassNode> classNodeMap = new HashMap<String, ClassNode>();
	
	private void analyzeClass(final ZipFile zipFile, final ZipEntry zipEntry) throws IOException {
		ClassNode cn =null;
		
		if (!classNodeMap.containsKey(zipEntry.getName().toLowerCase())) {
			InputStream is = zipFile.getInputStream(zipEntry);
			ClassReader classReader = new ClassReader(is);
			
			// create an empty ClassNode (in-memory representation of a class)
			final ClassNode classNode = new ClassNode();
			// have the ClassReader read the class file and populate the ClassNode with the corresponding information
			classReader.accept(classNode, 0);
			
			if(CallGraphBuilder.classSet == null || (CallGraphBuilder.classSet!=null && 
					CallGraphBuilder.classSet.contains(classNode.name))) {
				
				
				//classReader = new ClassReader(InlineJSR.inline(classNode, IOUtils.readFully(is, 1, true)));
				cn = new ClassNode();
				classReader.accept(cn, 0);
				classNodeMap.put(zipEntry.getName().toLowerCase(), cn);			
			}
		} else {
			cn = classNodeMap.get(zipEntry.getName().toLowerCase());
		}
		for (final ClassAnalyzer analyzer : analyzers) {
			if (cn!=null) {
				//System.out.println(cn.name);
				analyzer.analyze(zipFile.getName(), cn);
			}
		}
	}
	
	public byte[] getBytesFromFile(File file) throws IOException {
        InputStream is = new FileInputStream(file);
    
        // Get the size of the file
        long length = file.length();
    
        if (length > Integer.MAX_VALUE) {
            // File is too large
        }
    
        // Create the byte array to hold the data
        byte[] bytes = new byte[(int)length];
    
        // Read in the bytes
        int offset = 0;
        int numRead = 0;
        while (offset < bytes.length
               && (numRead=is.read(bytes, offset, bytes.length-offset)) >= 0) {
            offset += numRead;
        }
    
        // Ensure all the bytes have been read in
        if (offset < bytes.length) {
            throw new IOException("Could not completely read file "+file.getName());
        }
    
        // Close the input stream and return bytes
        is.close();
        return bytes;
    }
	
	private void analyzeClass(String name)  {
		ClassNode cn = null;
		if (!classNodeMap.containsKey(name)) {
			ClassReader classReader;
			try {
				File file = new File (name);
				FileInputStream fis = new FileInputStream(file);
				classReader = new ClassReader(fis);
				final ClassNode classNode = new ClassNode();
				// have the ClassReader read the class file and populate the ClassNode with the corresponding information
				classReader.accept(classNode, 0);
				if(CallGraphBuilder.classSet == null || (CallGraphBuilder.classSet!=null && 
						CallGraphBuilder.classSet.contains(classNode.name))) {
					
					classReader = new ClassReader(InlineJSR.inline(classNode, getBytesFromFile(file)));
					cn = new ClassNode();
					classReader.accept(cn, 0);
					classNodeMap.put(name, cn);			
				}
			} catch (IOException e) {
				System.out.println("weird but nor found " + name);
				e.printStackTrace();
			}
			// create an empty ClassNode (in-memory representation of a class)
		} else {
			cn = classNodeMap.get(name);
		}
		for (final ClassAnalyzer analyzer : analyzers) {
			if (cn!= null)
			analyzer.analyze(name, cn);
		}
	}

}
