package ch.unisi.inf.sp.type;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;

import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.MethodInsnNode;
import org.objectweb.asm.tree.MethodNode;

import usi.instrumentation.OpcodeReader;
import usi.types.instrumentation.LoopFinder;


public final class CallGraphBuilder implements ClassAnalyzer {

	private final ClassHierarchy hierarchy;
	
	
	public CallGraphBuilder(final ClassHierarchy hierarchy) {
		this.hierarchy = hierarchy;
	}
	
	public void analyze(final String location, final ClassNode classNode) {
		try {
			
			//here basic block builder can be plugged
			
			//System.out.println(classNode.name);
			
			final ClassType type = hierarchy.getOrCreateClass(classNode.name);
			final List<MethodNode> methodNodes = (List<MethodNode>)classNode.methods;
			for (final MethodNode methodNode : methodNodes) {
				
				
				
				final Subroutine subroutine = type.getSubroutine(methodNode.name, methodNode.desc);
				
				
				processLoopCallSite(type, classNode, methodNode, subroutine, type.getSubroutines());
				
				final InsnList instructions = methodNode.instructions;
				
				
				for (int i=0; i<instructions.size(); i++) {
					final AbstractInsnNode ins = instructions.get(i);
					
					if (ins instanceof MethodInsnNode) {
						getInnerSubroutine(subroutine, type.getSubroutines(), ins);
						
						if (inSubroutine) {
							/*...*/
						} else {
							innerRoutine = subroutine;
						}
						inSubroutine = false;
					} else {
						continue;
					}
				
					
					switch (ins.getOpcode()) {
					//case Opcodes.INVOKEDYNAMIC:
						// no idea what this is
						//break;
					case Opcodes.INVOKEINTERFACE:
						processVirtualOrInterfaceCallSite(type, innerRoutine, (MethodInsnNode)ins);
						break;
					case Opcodes.INVOKEVIRTUAL:
						processVirtualOrInterfaceCallSite(type, innerRoutine, (MethodInsnNode)ins);
						break;
					case Opcodes.INVOKESPECIAL:
						processSpecialCallSite(type, innerRoutine, (MethodInsnNode)ins);
						break;
					case Opcodes.INVOKESTATIC:
						processStaticCallSite(type, innerRoutine, (MethodInsnNode)ins);
						break;
					}
				}
			}
		} catch (final TypeInconsistencyException ex) {
			System.err.println(ex);
		}
	}
	
	private void processLoopCallSite(ClassType type, ClassNode cn, MethodNode mn, Subroutine method, Collection<Subroutine> subroutines) {
		for (Subroutine subroutine: subroutines) {
			if (!subroutine.isMethod()) {
				if (subroutine.getParent().equals(method)) {
					AbstractInsnNode ins = ((Loop)subroutine).getLoopBody().getHeader();
					
					final CallSite site = new CallSite(ins.getOpcode(), cn.name, subroutine.getName(), mn.desc);
					method.addCallSite(site);
					site.addPossibleTargetClass(type);
					processLoopCallSite(type, cn, mn, subroutine, subroutines);
				}
			}
		}
	}
	
	private Subroutine innerRoutine;
	private boolean inSubroutine = false;
	
	private void getInnerSubroutine (Subroutine method, Collection<Subroutine> subroutines, AbstractInsnNode ins) {
		for (Subroutine subroutine: subroutines) {
			if (subroutine.getParent()!= null) {
				if (subroutine.getParent().equals(method)) {
					//System.out.println("found loop in " + method.getName());
					/*
					 * should check to which subroutine this instruction belongs
					 * should find the most inner loop of that instruction
					 * and then add callsite to that subroutine
					 */
					if (((Loop)subroutine).getLoopBody().getInstructions().contains(ins)) {
						//should check in inner one and return that subroutine
						innerRoutine = subroutine;
						inSubroutine = true;
						getInnerSubroutine(subroutine, subroutines, ins);
						break;
					}
					
					
				}
			}
		}
		
	}

	private void resolveTargetInSuperClasses(final CallSite site, final ClassType type, final String methodName, final String methodDescriptor) {
		ClassType c = type;
		ClassType oldC = null;
		while (c!=null) {
			final Method method = (Method)c.getSubroutine(methodName, methodDescriptor);
			if (method!=null) {
				if (method.isAbstract()) {
					return; // no implementation anywhere up the tree
				} else {
					site.addPossibleTargetClass(c);
					return; // first concrete implementation found
				}
			}
			
			oldC = c;
			c = c.getSuperClass();
			if (c!=null && !c.equals(oldC)) {
				break;
			}
		}
		return; // no declaration anywhere. bug!
	}

	private void resolveTargetInSubClasses(final CallSite site, final ClassType type, final String methodName, final String methodDescriptor) {
		final Method method = (Method)type.getSubroutine(methodName, methodDescriptor);
		if (method!=null && !method.isAbstract()) {
			site.addPossibleTargetClass(type);
		}
		for (final ClassType subType : type.getSubTypes()) {
			if (!type.equals(subType) && !subType.getSubTypes().contains(type)) {
				
				resolveTargetInSubClasses(site, subType, methodName, methodDescriptor);
			}
		}
	}
	
	private void processVirtualOrInterfaceCallSite(final ClassType type, final Subroutine subroutine, final MethodInsnNode ins) {
		final CallSite site = new CallSite(ins.getOpcode(), ins.owner, ins.name, ins.desc);
		subroutine.addCallSite(site);
		try {
			final ClassType targetClass = hierarchy.getOrCreateClass(ins.owner);
			resolveTargetInSuperClasses(site, targetClass, ins.name, ins.desc);
			resolveTargetInSubClasses(site, targetClass, ins.name, ins.desc);
		} catch (final TypeInconsistencyException ex) {
			System.err.println("Expecting a class or interface");
		}
	}
	
	private void processSpecialCallSite(final ClassType type, final Subroutine subroutine, final MethodInsnNode ins) {
		final CallSite site = new CallSite(ins.getOpcode(), ins.owner, ins.name, ins.desc);
		subroutine.addCallSite(site);
		try {
			final ClassType targetClass = hierarchy.getOrCreateClass(ins.owner);
			resolveTargetInSuperClasses(site, targetClass, ins.name, ins.desc);
		} catch (final TypeInconsistencyException ex) {
			System.err.println("Expecting a class or interface");
		}
	}
	
	private void processStaticCallSite(final ClassType type, final Subroutine subroutine, final MethodInsnNode ins) {
		final CallSite site = new CallSite(ins.getOpcode(), ins.owner, ins.name, ins.desc);
		subroutine.addCallSite(site);
		try {
			final ClassType targetClass = hierarchy.getOrCreateClass(ins.owner);
			final Method targetMethod = (Method)targetClass.getSubroutine(ins.name, ins.desc);
			if (targetMethod!=null) {
				site.addPossibleTargetClass(targetClass);
			}
		} catch (final TypeInconsistencyException ex) {
			System.err.println("Expecting a class or interface");
		}
	}

	public static String[] fixArgs(String[] args) {
		String[] newArgs;
		int global = 0;
		for (int i = 1; i<args.length; i++) {
			if (!args[i].endsWith(".zip") && 
					(!args[i].endsWith(".jar")) &&
					(!args[i].endsWith(".class"))) {
				//we should concatenate next args until it will finnish
				
				StringBuffer sb = new StringBuffer();
				//sb.append("\"");
				sb.append(args[i]);
				int count = 0;
				for (int j = i+1; j<args.length; j++) {
					if (!args[j].endsWith(".zip") && 
							(!args[j].endsWith(".jar")) &&
							(!args[j].endsWith(".class"))) {
						sb.append(" ");
						sb.append(args[j]);
						count++;
					} else {
						count++;
						sb.append(" ");
						sb.append(args[j]);
						//sb.append("\"");
						
						args[i] = sb.toString();
						//System.out.println(args[i]);
						i = j;
						global = global + count;
						break;
					}
				}
			}
		}
		/*newArgs = new String[args.length-global];

		newArgs[0] = args[0];
		int number = 1;
		for (int i = 1; i<args.length; i++) {
			if ((args[i].endsWith(".zip") || 
					(args[i].endsWith(".jar")) ||
					(args[i].endsWith(".class"))) &&
					(args[i].contains(" "))) {
				
				newArgs[number] = args[i];
				number++;
			}
		}*/
		
		return args;
	}
	
	public static HashSet<String> classSet;

	
	private static void fillSet(String name) throws IOException {
		
		File file = new File("/Users/hauswirm/app/QualitasCorpus-20100719r/Systems/" + name);
		for (File f: file.listFiles()) {
			if (!f.getName().startsWith(".")) {
				
				File txt = new File (f+"/metadata/contents.txt");
				BufferedReader br = new BufferedReader(new FileReader(txt));
				
				String line;
				
				while ((line = br.readLine())!= null) {
					if (!line.startsWith("#") && !line.equals("")) {
						String[] lines = line.split("\t");
						//System.out.println(lines[0].replaceAll("\\.", "/"));
						if (lines[3].equals("0"))
							//System.out.println(lines[0]);
							classSet.add(lines[0].replaceAll("\\.", "/"));
					}
				}
				
				break;
			}
		}
		
	}
	
	public static void main(String[] args) throws IOException {
		
		//only for qualitas
		if (args[0].endsWith("M.txt")) {
			classSet = new HashSet<String>();
			fillSet(args[0].substring(0, args[0].lastIndexOf("M.txt")));
		}
		
		ArchiveScanner scanner = new ArchiveScanner();
		final ClassHierarchyBuilder classHierarchyBuilder = new ClassHierarchyBuilder();
		scanner.addAnalyzer(classHierarchyBuilder);
		
		args = fixArgs(args);
		//for (final String arg : args) {
		 System.out.print(args[0]);
		//}
		for (int i=1; i<args.length; i++) {
			scanner.scan(args[i]);
		}
		scanner.removeAnalyzer(classHierarchyBuilder);
		
		CallGraphBuilder callGraphBuilder = new CallGraphBuilder(classHierarchyBuilder.getClassHierarchy());
		scanner.addAnalyzer(callGraphBuilder);
		for (int i=1; i<args.length; i++) {
			scanner.scan(args[i]);
		}
		scanner = null;
		callGraphBuilder = null;
		
		final CallRoutineDumper rdumper = new CallRoutineDumper(args[0]);
		rdumper.constructTree(classHierarchyBuilder.getClassHierarchy());
		
		/*final CallGraphDumper dumper = new CallGraphDumper();
		dumper.dumpDot(classHierarchyBuilder.getClassHierarchy(), "HierarchyWithMethods.dot", false, false);
		dumper.dumpDot(classHierarchyBuilder.getClassHierarchy(), "HierarchyWithMethodsWithDeclaredCallEdges.dot", true, false);
		dumper.dumpDot(classHierarchyBuilder.getClassHierarchy(), "HierarchyWithMethodsWithPossibleCallEdges.dot", false, true);
		dumper.dumpDot(classHierarchyBuilder.getClassHierarchy(), "HierarchyWithMethodsWithCallEdges.dot", true, true);
		
		final ClassHierarchyWithCallEdgeDumper dumper2 = new ClassHierarchyWithCallEdgeDumper();
		dumper2.dumpDot(classHierarchyBuilder.getClassHierarchy(), "Hierarchy.dot", false, false);
		dumper2.dumpDot(classHierarchyBuilder.getClassHierarchy(), "HierarchyWithDeclaredCallEdges.dot", true, false);
		dumper2.dumpDot(classHierarchyBuilder.getClassHierarchy(), "HierarchyWithPossibleCallEdges.dot", false, true);
		dumper2.dumpDot(classHierarchyBuilder.getClassHierarchy(), "HierarchyWithCallEdges.dot", true, true);
	*/
	}
}
