package ch.unisi.inf.sp.type;

import java.io.IOException;
import java.util.List;

import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.MethodNode;

import usi.types.instrumentation.LoopFinder;


public final class ClassHierarchyBuilder implements ClassAnalyzer {

	private final ClassHierarchy classHierarchy;
	
	
	public ClassHierarchyBuilder() {
		this.classHierarchy = new ClassHierarchy();
	}
	
	public ClassHierarchy getClassHierarchy() {
		return classHierarchy;
	}
	
	public void analyze(final String location, final ClassNode clazz) {
		try {
			final ClassType classType = classHierarchy.getOrCreateClass(clazz.name);
			if (classType.isResolved()) {
				//System.err.println("WARNING: Class "+classType.getInternalName()+" defined multiple times");
				return;
			}
			classType.setLocation(location);
			// modifiers
			classType.setModifiers(clazz.access);
			// super class
			final ClassType superClassType = classHierarchy.getOrCreateClass(clazz.superName);
			classType.setSuperClass(superClassType);
			// interfaces
			final List<String> interfaces = (List<String>)clazz.interfaces;
			for (final String interfaceName : interfaces) {
				final ClassType interfaceType = classHierarchy.getOrCreateClass(interfaceName);
				classType.addInterface(interfaceType);
			}
			// methods
			final List<MethodNode> methodNodes = (List<MethodNode>)clazz.methods;
			for (final MethodNode methodNode : methodNodes) {
				final Method method = new Method(clazz.name, methodNode.name, methodNode.desc, methodNode.access);
				classType.addSubroutine(method);
				
				LoopFinder loopFinder = new LoopFinder();
				List<ch.unisi.inf.sp.type.LoopConstruct> loops = loopFinder.findLoops(clazz, methodNode);
				
				method.numberOgLoops = loops.size();
				
				
				for (ch.unisi.inf.sp.type.LoopConstruct loopConstr: loops) {
					
					final Loop loop;
					if (loopConstr.getOwner()==null) {
						loop = new Loop(clazz.name, methodNode.name + "-" + methodNode.instructions.indexOf(loopConstr.getHeader()), methodNode.desc, methodNode.access, method, loopConstr);
						classType.addSubroutine(loop);
					} 
					
				}
				for (ch.unisi.inf.sp.type.LoopConstruct loopConstr: loops) {
					
					recursivelyCreateLoopNodes(classType, loopConstr, methodNode, clazz.name); 
					
				}
			}
			classType.setResolved();
		} catch (final TypeInconsistencyException ex) {
			System.err.println(ex);
		}
	}
	
	private void recursivelyCreateLoopNodes (ClassType classType, ch.unisi.inf.sp.type.LoopConstruct loopConstr,
			MethodNode methodNode, String className) {
					
		
		final Loop loop;
		if (loopConstr.getOwner()!=null) {
			if (classType.getSubroutine(methodNode.name + "-" + methodNode.instructions.indexOf(loopConstr.getHeader()), methodNode.desc)==null) {
				//it means that we didn't create the owner loop of that loop, so we should recursively do that
				if (classType.getSubroutine(methodNode.name+"-"+methodNode.instructions.indexOf(loopConstr.getOwner().getHeader()), methodNode.desc)==null) {
					recursivelyCreateLoopNodes(classType, loopConstr.getOwner(), methodNode, className);
					
				}
				
				loop = new Loop
				(className, methodNode.name+"-" + methodNode.instructions.indexOf(loopConstr.getHeader()), methodNode.desc, methodNode.access,
						classType.getSubroutine(methodNode.name+"-"+methodNode.instructions.indexOf(loopConstr.getOwner().getHeader()), methodNode.desc), loopConstr);
				classType.addSubroutine(loop);
				
			} 
		} 
	}

	public static void main(final String[] args) throws IOException {
		final ArchiveScanner scanner = new ArchiveScanner();
		final ClassHierarchyBuilder builder = new ClassHierarchyBuilder();
		scanner.addAnalyzer(builder);
		for (int i=0; i<args.length; i++) {
			scanner.scan(args[i]);
		}
		final ClassHierarchyDumper d = new ClassHierarchyDumper();
		d.dump(builder.getClassHierarchy());
		d.dumpDot(builder.getClassHierarchy());
		final StatisticsDumper s = new StatisticsDumper();
		s.dump(builder.getClassHierarchy());
	}
	
}
