package ch.unisi.inf.sp.type;

import java.io.IOException;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.MethodInsnNode;
import org.objectweb.asm.tree.MethodNode;


public final class PolymorphicCallSiteAnalyzer implements ClassAnalyzer {

	private static class ArityHistogram {
		private int[] bins;
		public ArityHistogram() {
			bins = new int[1];
		}
		public void count(int arity) {
			if (arity>=bins.length) {
				//grow
				int[] newBins = new int[arity+1];
				System.arraycopy(bins, 0, newBins, 0, bins.length);
				bins = newBins;
			}
			bins[arity]++;
		}
		public void dump() {
			System.out.println("Arity\tSites");
			for (int i=1; i<bins.length; i++) {
				System.out.println(i+"\t"+bins[i]);
			}
		}
	}
	private final ClassHierarchy hierarchy;
	private int dynamicCallSites;
	private int interfaceCallSites;
	private int specialCallSites;
	private int staticCallSites;
	private int virtualCallSites;
	private int ppSitesWithFinalTargetType;
	private int ppSitesWithFinalTargetMethod;
	private int ppSitesWithFinalTarget;
	private int ppSites;
	private int ppCallTargetArity;
	private ArityHistogram histogram;
	
	
	public PolymorphicCallSiteAnalyzer(final ClassHierarchy hierarchy) {
		this.hierarchy = hierarchy;
		histogram = new ArityHistogram();
	}
	
	public void analyze(final String location, final ClassNode classNode) {
		final List<MethodNode> methods = (List<MethodNode>)classNode.methods;
		for (final MethodNode method : methods) {
			final InsnList instructions = method.instructions;
			for (int i=0; i<instructions.size(); i++) {
				final AbstractInsnNode ins = instructions.get(i);
				switch (ins.getOpcode()) {
				case Opcodes.INVOKEDYNAMIC:
					dynamicCallSites++;
					break;
				case Opcodes.INVOKEINTERFACE:
					interfaceCallSites++;
					analyze((MethodInsnNode)ins);
					break;
				case Opcodes.INVOKEVIRTUAL:
					virtualCallSites++;
					analyze((MethodInsnNode)ins);
					break;
				case Opcodes.INVOKESPECIAL:
					specialCallSites++;
					break;
				case Opcodes.INVOKESTATIC:
					staticCallSites++;
					break;
				}
			}
		}
	}
	
	/**
	 * Given a class and a method name+descriptor, find the closest super class (starting with the given class itself)
	 * that implements this method to determine whether the method is final.
	 */
	private boolean isMethodFinal(final ClassType type, final String methodName, final String methodDescriptor) {
		ClassType c = type;
		while (c!=null) {
			final Method method = (Method)c.getSubroutine(methodName, methodDescriptor);
			if (method!=null) {
				if (method.isFinal()) {
					return true; // it's final
				}
				if (method.isAbstract()) {
					return false; // can't possibly have a superclass implementing it
				}
			}
			c = c.getSuperClass();
		}
		return false; // we searched all the way to java.lang.Object, without the method being final anywhere
	}

	private ClassType findTargetTypeInSuperClasses(final ClassType type, final String methodName, final String methodDescriptor) {
		ClassType c = type;
		while (c!=null) {
			final Subroutine subroutine = c.getSubroutine(methodName, methodDescriptor);
			if (subroutine!=null) {
				if (subroutine.isAbstract()) {
					return null; // no implementation anywhere up the tree
				} else {
					return c; // first concrete implementation found
				}
			}
			c = c.getSuperClass();
		}
		return null; // no declaration anywhere. bug!
	}

	private void collectPotentialCallTargetTypes(final ClassType type, final String methodName, final String methodDescriptor, final Set<ClassType> targetTypes) {
		final Subroutine subroutine = type.getSubroutine(methodName, methodDescriptor);
		if (subroutine!=null && !subroutine.isAbstract()) {
			targetTypes.add(type);
		}
		for (final ClassType subType : type.getSubTypes()) {
			collectPotentialCallTargetTypes(subType, methodName, methodDescriptor, targetTypes);
		}
	}
	
	private Set<ClassType> findPotentialCallTargetTypes(final ClassType type, final String methodName, final String methodDescriptor) {
		final Set<ClassType> targetTypes = new HashSet<ClassType>();
		// traversal up the tree, including type
		final ClassType superImplementation = findTargetTypeInSuperClasses(type, methodName, methodDescriptor);
		if (superImplementation!=null) {
			targetTypes.add(superImplementation);
		}
		// traversal down the tree, also including type (type may be added to targetTypes twice, but it's a set, so that's ok)
		collectPotentialCallTargetTypes(type, methodName, methodDescriptor, targetTypes);
		return targetTypes;
	}
	
	private void analyze(final MethodInsnNode ins) {
		try {
			final ClassType targetClass = hierarchy.getOrCreateClass(ins.owner);
			if (targetClass.isFinal()) {
				ppSitesWithFinalTargetType++;
			}
			final boolean targetMethodIsFinal = isMethodFinal(targetClass, ins.name, ins.desc);
			if (targetMethodIsFinal) {
				ppSitesWithFinalTargetMethod++;				
			}
			if (targetClass.isFinal() || targetMethodIsFinal) {
				ppSitesWithFinalTarget++;
			}
			final int arity = findPotentialCallTargetTypes(targetClass, ins.name, ins.desc).size();
			if (arity==0) {
				//System.out.println("  No targets for: "+ins.owner+"."+ins.name+ins.desc);
			} else {
				ppCallTargetArity += arity;
				ppSites++;
				histogram.count(arity);
				if (arity>10) {
					System.out.println("  "+arity+" targets for: "+ins.owner+"."+ins.name+ins.desc);
				}
			}
		} catch (final TypeInconsistencyException ex) {
			System.err.println("Expecting a class or interface");
		}
	}
	
	public int getDynamicCallSiteCount() {
		return dynamicCallSites;
	}
	
	public int getInterfaceCallSiteCount() {
		return interfaceCallSites;
	}
	
	public int getVirtualCallSiteCount() {
		return virtualCallSites;
	}
	
	public int getSpecialCallSiteCount() {
		return specialCallSites;
	}
	
	public int getStaticCallSiteCount() {
		return staticCallSites;
	}
	
	public int getPotentiallyPolymorphicCallSitesWithFinalTargetType() {
		return ppSitesWithFinalTargetType;
	}
	
	public int getPotentiallyPolymorphicCallSitesWithFinalTargetMethod() {
		return ppSitesWithFinalTargetMethod;
	}
	
	public int getPotentiallyPolymorphicCallSitesWithFinalTarget() {
		return ppSitesWithFinalTarget;
	}
	
	public int getResolvablePotentiallyPolymorphicCallSites() {
		return ppSites;
	}
	
	public int getTotalTargetCountOfResolvablePotentiallyPolymorphicCallSites() {
		return ppCallTargetArity;
	}
	

	public static void main(final String[] args) throws IOException {
		for (final String arg : args) {
			System.out.println(arg);
		}
		final ArchiveScanner scanner = new ArchiveScanner();
		final ClassHierarchyBuilder builder = new ClassHierarchyBuilder();
		scanner.addAnalyzer(builder);
		for (int i=0; i<args.length; i++) {
			scanner.scan(args[i]);
		}
		scanner.removeAnalyzer(builder);
		
		final PolymorphicCallSiteAnalyzer callCounter = new PolymorphicCallSiteAnalyzer(builder.getClassHierarchy());
		scanner.addAnalyzer(callCounter);
		for (int i=0; i<args.length; i++) {
			scanner.scan(args[i]);
		}
		System.out.println("Static count of call sites: ");
		System.out.println(" dynamic:\t"+callCounter.getDynamicCallSiteCount());
		System.out.println(" interface:\t"+callCounter.getInterfaceCallSiteCount());
		System.out.println(" virtual:\t"+callCounter.getVirtualCallSiteCount());
		System.out.println(" special:\t"+callCounter.getSpecialCallSiteCount());
		System.out.println(" static:\t"+callCounter.getStaticCallSiteCount());
		final int poly = callCounter.getInterfaceCallSiteCount()+callCounter.getVirtualCallSiteCount();
		final int all = poly+callCounter.getSpecialCallSiteCount()+callCounter.getStaticCallSiteCount();
		System.out.println("potentially polymorphic call sites: "+poly+" ("+(poly*100/all)+"%)");
		final int ppFT = callCounter.getPotentiallyPolymorphicCallSitesWithFinalTargetType();
		System.out.println("potentially polymorphic call sites with final target type: "+ppFT+" ("+(ppFT*100/all)+"%)");
		final int ppFM = callCounter.getPotentiallyPolymorphicCallSitesWithFinalTargetMethod();
		System.out.println("potentially polymorphic call sites with final target method: "+ppFM+" ("+(ppFM*100/all)+"%)");
		final int ppF = callCounter.getPotentiallyPolymorphicCallSitesWithFinalTarget();
		System.out.println("potentially polymorphic call sites with final target: "+ppF+" ("+(ppF*100/all)+"%)");
		final int ppCount = callCounter.getResolvablePotentiallyPolymorphicCallSites();
		System.out.println("  resolvable pp sites: "+ppCount+" ("+(ppCount*100/all)+"%)");
		final int ppTotalArity = callCounter.getTotalTargetCountOfResolvablePotentiallyPolymorphicCallSites();
		System.out.println("  total targets of resolvable pp sites: "+ppTotalArity);
		System.out.println("  targets/resolvable pp sites: "+ppTotalArity/(double)ppCount);
		callCounter.histogram.dump();
	}
}
