package ch.usi.inf.sape.graph;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;


/**
 * A FlowGraph is a directed graph of Nodes and Edges, 
 * with a dedicated entry Node from which all other nodes are reachable.
 * 
 * @author Matthias.Hauswirth@usi.ch
 */
public class FlowGraph implements Cloneable {

	public static final boolean DEBUG = true;
	
	
	// seal the graph (e.g. after construction) to prevent any modifications
	private boolean sealed;
	private Node entryNode;
	private ArrayList<Node> nodes;
	private ArrayList<Edge> edges;
	
	
	/**
	 * Construct an empty FlowGraph (no nodes, no edges, no node set as entryNode).
	 */
	public FlowGraph() {
		nodes = new ArrayList<Node>();
		edges = new ArrayList<Edge>();
	}
	
	/**
	 * Construct a deep copy of the given FlowGraph.
	 * 
	 * If you subclass FlowGraph, you need to override this method
	 * to properly clone the attributes in your class.
	 * From your subclass, call super.clone() to get the clone,
	 * then clone all your subclass-specific mutable fields.
	 */
	public FlowGraph clone() throws CloneNotSupportedException {
		final FlowGraph clone = (FlowGraph)super.clone();
		clone.nodes = new ArrayList<Node>();
		clone.edges = new ArrayList<Edge>();
		// copy structure (nodes and edges) from original
		final HashMap<Node,Node> nodeMap = new HashMap<Node,Node>();
		for (final Node node : getNodes()) {
			final Node clonedNode = (Node)node.clone();
			nodeMap.put(node, clonedNode);
			clone.addNode(clonedNode);
		}
		for (final Edge edge : getEdges()) {
			final Edge clonedEdge = (Edge)edge.clone();
			clonedEdge.setSource(nodeMap.get(edge.getSource()));
			clonedEdge.setTarget(nodeMap.get(edge.getTarget()));
			clone.addEdge(clonedEdge);
		}
		clone.makeEntry(nodeMap.get(getEntryNode()));
		return clone;
	}

	public void check(final boolean terminateOnProblem) {
		if (nodes==null) {
			warn("nodes==null", terminateOnProblem);
		}
		if (edges==null) {
			warn("edges==null", terminateOnProblem);
		}
		if (edges.size()>0 && nodes.size()==0) {
			warn("edges but no nodes", terminateOnProblem);
		}
		for (final Node node : getNodes()) {
			if (node.getInEdges()==null) {
				warn("node "+node+" has no inEdges", terminateOnProblem);
			}
			if (node.getOutEdges()==null) {
				warn("node "+node+" has no outEdges", terminateOnProblem);
			}
		}
		for (final Edge edge : getEdges()) {
			if (edge.getSource()==null) {
				warn("edge "+edge.getSource()+" -> "+edge.getTarget()+" has null source", terminateOnProblem);
			}
			if (edge.getTarget()==null) {
				warn("edge "+edge.getSource()+" -> "+edge.getTarget()+" has null target", terminateOnProblem);
			}
		}
		for (final Edge edge : getEdges()) {
			if (!nodes.contains(edge.getSource())) {
				warn("edge "+edge.getSource()+" -> "+edge.getTarget()+"'s source is not a node in this graph", terminateOnProblem);
			}
			if (!nodes.contains(edge.getTarget())) {
				warn("edge "+edge.getSource()+" -> "+edge.getTarget()+"'s target is not a node in this graph", terminateOnProblem);
			}
		}
		for (final Edge edge : getEdges()) {
			if (!edge.getSource().getOutEdges().contains(edge)) {
				warn("edge "+edge.getSource()+" -> "+edge.getTarget()+"'s source's outEdges do not contain edge", terminateOnProblem);
			}
			if (!edge.getTarget().getInEdges().contains(edge)) {
				warn("edge "+edge.getSource()+" -> "+edge.getTarget()+"'s target's inEdges do not contain edge", terminateOnProblem);
			}
		}
	}

	/**
	 * Print to stderr all nodes that are unreachable.
	 * Note: This will clear all marks (so don't call it from the middle of an algorithm that marks nodes!)
	 */
	public void checkReachability(final boolean terminateOnProblem) {
		clearAllMarks();
		if (entryNode!=null) {
			dfsMark(entryNode);			
		}
		for (final Node n : nodes) {
			if (!isMarked(n)) {
				warn("Unreachable node: "+n, terminateOnProblem);
			}
		}
	}
	
	public List<Node> getUnreachableNodes() {
		List<Node> l = new ArrayList<Node>();
		clearAllMarks();
		if (entryNode!=null) {
			dfsMark(entryNode);			
		}
		for (final Node n : nodes) {
			if (!isMarked(n)) {
				l.add(n);
			}
		}
		return l;
	}
	
	/**
	 * Collections are based on Object.equals(), not ==.
	 * This method tests whether a collection already contains the given element
	 * (not just an equal element, but the specific object).
	 * 
	 * @param <T>
	 * @param c
	 * @param e
	 * @return
	 */
	private static <T> boolean containsSame(final Collection<T> c, final T e) {
		final Iterator<T> it = c.iterator();
		while (it.hasNext()) {
			if (it.next()==e) {
				return true;
			}
		}
		return false;
	}
	
	private void dfsMark(final Node node) {
		if (isMarked(node)) {
			return;
		} else {
			mark(node);
			for (final Edge edge : node.getOutEdges()) {
				dfsMark(edge.getTarget());
			}
		}
	}
	
	private static void warn(final String message, final boolean terminate) {
		if (terminate) {
			throw new IllegalStateException(message);
		} else {
			System.err.println(message);
		}
	}
	
	public boolean hasParallelEdges() {
		for (final Edge edge1 : getEdges()) {
			for (final Edge edge2 : getEdges()) {
				if (edge1!=edge2 && edge1.getSource()==edge2.getSource() && edge1.getTarget()==edge2.getTarget()) {
					return true;
				}
			}
		}
		return false;
	}
	
	public boolean hasSelfLoops() {
		for (final Edge edge : getEdges()) {
			if (edge.getSource()==edge.getTarget()) {
				return true;
			}
		}
		return false;
	}
	
	
	//--- modifications
	public void seal() {
		sealed = true;
	}
	
	public void unseal() {
		sealed = false;
	}
	
	public void makeEntry(final Node node) {
		if (DEBUG) {
			if (sealed) {
				throw new IllegalStateException("Graph is sealed!");
			}
			if (node==null) {
				throw new IllegalArgumentException("Given node must not be null!");
			}
			if (!containsSame(nodes, node)) {
				throw new IllegalStateException("Given node must already be in graph!");
			}
		}
		entryNode = node;
	}
	
	public void addNode(final Node node) {
		if (DEBUG) {
			if (sealed) {
				throw new IllegalStateException("Graph is sealed!");
			}
			if (node==null) {
				throw new IllegalArgumentException("Given node must not be null!");
			}
			if (nodes.contains(node)) {
				throw new IllegalArgumentException("Given node must not yet be in the graph!");			
			}
			if (node.getInEdges().size()>0) {
				throw new IllegalArgumentException("Given node must not yet have inEdges!");
			}
			if (node.getOutEdges().size()>0) {
				throw new IllegalArgumentException("Given node must not yet have outEdges!");
			}
		}
		nodes.add(node);
	}
	
	public void removeNode(final Node node) {
		if (DEBUG) {
			if (sealed) {
				throw new IllegalStateException("Graph is sealed!");
			}
			if (node==null) {
				throw new IllegalArgumentException("Given node must not be null!");
			}
			if (!containsSame(nodes, node)) {
				throw new IllegalArgumentException("Given node must still be in the graph!");			
			}
			if (node.getInEdges().size()>0) {
				throw new IllegalArgumentException("Given node must not have inEdges left!");
			}
			if (node.getOutEdges().size()>0) {
				throw new IllegalArgumentException("Given node must not have outEdges left!");
			}
		}
		nodes.remove(node);
	}
	
	public void addEdge(final Edge edge) {
		if (DEBUG) {
			if (sealed) {
				throw new IllegalStateException("Graph is sealed!");
			}
			if (edge==null) {
				throw new IllegalArgumentException("Given edge must not be null!");
			}
			if (edges.contains(edge)) {
				throw new IllegalArgumentException("Given edge must not yet be in the graph!");			
			}
			if (!containsSame(nodes, edge.getSource())) {
				throw new IllegalArgumentException("Given edge must have a source that is already in the graph!");			
			}
			if (!containsSame(nodes, edge.getTarget())) {
				throw new IllegalArgumentException("Given edge must have a target that is already in the graph!");			
			}
		}
		edges.add(edge);
		edge.getSource().addOutEdge(edge);
		edge.getTarget().addInEdge(edge);
	}
	
	public void removeEdge(final Edge edge) {
		if (DEBUG) {
			if (sealed) {
				throw new IllegalStateException("Graph is sealed!");
			}
			if (edge==null) {
				throw new IllegalArgumentException("Given edge must not be null!");
			}
			if (!containsSame(edges, edge)) {
				throw new IllegalArgumentException("Given edge must still be in the graph!");			
			}
			if (!containsSame(edge.getSource().getOutEdges(), edge)) {
				throw new IllegalArgumentException("Source of given edge must still contain the edge as an outEdge");
			}
			if (!containsSame(edge.getTarget().getInEdges(), edge)) {
				throw new IllegalArgumentException("Target of given edge must still contain the edge as an inEdge");
			}
		}
		edges.remove(edge);
		edge.getSource().removeOutEdge(edge);
		edge.getTarget().removeInEdge(edge);
	}
	
	
	//--- properties
	public void clearAllMarks() {
		for (final Node node : nodes) {
			node.clearMark();
		}
	}
	
	public void mark(final Node node) {
		node.mark();
	}
	
	public boolean isMarked(final Node node) {
		return node.isMarked();
	}
	
	
	//--- queries
	public Node getEntryNode() {
		return entryNode;
	}
	
	public List<Node> getNodes() {
		if (DEBUG) {
			return Collections.unmodifiableList(nodes);
		} else {
			return nodes;
		}
	}
	
	public List<Edge> getEdges() {
		if (DEBUG) {
			return Collections.unmodifiableList(edges);
		} else {
			return edges;
		}
	}
	
}
