package ch.usi.inf.sape.graph;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;


/**
 * Don't modify Nodes. Go through the Graph (e.g. to mark or to add/remove in/out edges).
 * 
 * @author Matthias.Hauswirth@usi.ch
 */
public class Node implements Cloneable, Serializable {

	private ArrayList<Edge> outEdges;
	private ArrayList<Edge> inEdges;
	private boolean marked;
	
	
	public Node() {
		outEdges = new ArrayList<Edge>();
		inEdges = new ArrayList<Edge>();
		marked = false;
	}
	
	/**
	 * Create a clone of the node,
	 * but the clone does NOT contain any inEdges and outEdges!
	 * 
	 * The FlowGraph, when it is being cloned, will establish
	 * inEdges, outEdges during its cloning process.
	 * 
	 * You need to override this method if you implement subclasses of Node 
	 * and want to properly support cloning a FlowGraph.
	 */
	protected Node clone() throws CloneNotSupportedException {
		final Node clone = (Node)super.clone();
		clone.outEdges = new ArrayList<Edge>();
		clone.inEdges = new ArrayList<Edge>();
		return clone;
	}
	
	
	//--- marking (don't call these methods directly, go through Graph)
	public boolean isMarked() {
		return marked;
	}
	
	public void mark() {
		marked = true;
	}
	
	public void clearMark() {
		marked = false;
	}
	
	
	//--- queries
	public List<Edge> getOutEdges() {
		if (FlowGraph.DEBUG) {
			return Collections.unmodifiableList(outEdges);
		} else {
			return outEdges;
		}
	}
	
	public List<Edge> getInEdges() {
		if (FlowGraph.DEBUG){
			return Collections.unmodifiableList(inEdges);
		} else {
			return inEdges;
		}
	}
	
	
	//--- update (don't call these methods directly, go through Graph)
	void addOutEdge(final Edge edge) {
		outEdges.add(edge);
	}
	
	void addInEdge(final Edge edge) {
		inEdges.add(edge);
	}
	
	void removeOutEdge(final Edge edge) {
		outEdges.remove(edge);
	}
	
	void removeInEdge(final Edge edge) {
		inEdges.remove(edge);
	}
	
}
