package ch.usi.inf.sape.graph.algorithms;

import java.util.ArrayList;

import ch.usi.inf.sape.graph.Edge;
import ch.usi.inf.sape.graph.Factory;
import ch.usi.inf.sape.graph.FlowGraph;
import ch.usi.inf.sape.graph.Node;


/**
 * Implementation of Hecht/Ullman "Flow Graph Reducibility" test.
 * 
 * @see Hecht'72: http://doi.acm.org/10.1145/800152.804919
 *
 * @author Matthias.Hauswirth@usi.ch
 */
public final class HechtUllmanReducibilityTest {
	
	/**
	 * Test whether the given graph is reducible.
	 * 
	 * Note that this method will CHANGE the given graph!
	 * It will turn either into a single-node graph (if it is reducible),
	 * or into a graph that cannot be reduced any further.
	 * 
	 * If you don't want the graph to be changed, run this algorithm on a clone of the graph.
	 * 
	 * @throws CloneNotSupportedException We use Edge.clone() when reducing the graph, 
	 * this will throw an exception if the concrete Edge subclass does not support cloning.
	 */
	public static boolean isReducible(final FlowGraph g, final Factory factory) throws CloneNotSupportedException {
		ParallelEdgeCollapser.eliminateParallelEdges(g);
		eliminateUnreachableNodes(g);
		boolean changed = true;
		while (changed) {
			eliminateCycles(g);
			changed = collapseVertices(g, factory);
		}
		return hasSingleNode(g);
	}
	
	/**
	 * Some nodes, e.g. in ASM CFGs, are unreachable.
	 * To test the reducibility of those graphs, we have to first remove any unreachable nodes.
	 * @param g FlowGraph
	 */
	private static void eliminateUnreachableNodes(final FlowGraph g) {
		g.clearAllMarks();
		markReachableNodes(g, g.getEntryNode());
		final ArrayList<Node> unreachableNodes = new ArrayList<Node>();
		for (final Node node : g.getNodes()) {
			if (!node.isMarked()) {
				unreachableNodes.add(node);
			}
		}
		for (final Node node : unreachableNodes) {
			removeUnreachableNode(g, node);
		}
	}
	
	private static void markReachableNodes(final FlowGraph g, final Node n) {
		g.mark(n);
		for (final Edge edge : n.getOutEdges()) {
			final Node target = edge.getTarget();
			if (!target.isMarked()) {
				markReachableNodes(g, target);
			}
		}
	}
	
	private static void removeUnreachableNode(final FlowGraph g, final Node n) {
		// clone, because we will modify the graph during traversal
		final ArrayList<Edge> inEdges = new ArrayList<Edge>(n.getInEdges());
		for (final Edge edge : inEdges) {
			g.removeEdge(edge);
		}
		// clone, because we will modify the graph during traversal
		final ArrayList<Edge> outEdges = new ArrayList<Edge>(n.getOutEdges());
		for (final Edge edge : outEdges) {
			g.removeEdge(edge);
		}
		g.removeNode(n);
	}
	
	/**
	 * Performs a depth-first search, identifying and eliminating cycles (self-loops) at each node.
	 * @param g the Graph to traverse
	 * @return whether it changed the graph (whether it found any cycles)
	 */
	private static boolean eliminateCycles(final FlowGraph g) {
		g.clearAllMarks();
		return eliminateCycles(g, g.getEntryNode());
	}
	
	private static boolean eliminateCycles(final FlowGraph g, final Node b) {
		boolean changed = false;
		g.mark(b);
		// clone, because we will modify the graph during traversal
		final ArrayList<Edge> outEdges = new ArrayList<Edge>(b.getOutEdges());
		for (final Edge e : outEdges) {
			final Node target = e.getTarget();
			if (target==b) {
				g.removeEdge(e);
				changed = true;
			} else {
				if (!g.isMarked(target)) {
					changed |= eliminateCycles(g, target);
				}
			}
		}
		return changed;
	}
	
	private static boolean collapseVertices(final FlowGraph g, final Factory factory) throws CloneNotSupportedException {
		boolean changed = false;
		g.clearAllMarks();
		final Node entry = g.getEntryNode();
		g.mark(entry);
		// clone, because we will modify the graph during traversal
		final ArrayList<Edge> outEdges = new ArrayList<Edge>(entry.getOutEdges());
		for (final Edge edge : outEdges) {
			final Node target = edge.getTarget();
			changed |= collapseVertices(g, factory, entry, target);
		}
		return changed;
	}

	private static boolean collapseVertices(final FlowGraph g, final Factory factory, final Node parent, final Node child) throws CloneNotSupportedException {
		boolean changed = false;
		// first, recursively traverse
		g.mark(child);
		// clone, because we will modify the graph during traversal
		final ArrayList<Edge> outEdges = new ArrayList<Edge>(child.getOutEdges());
		for (final Edge edge : outEdges) {
			final Node target = edge.getTarget();
			if (!target.isMarked()) {
				changed |= collapseVertices(g, factory, child, target);
			}
		}
		// then collapse this edge, if necessary
		if (child.getInEdges().size()==1) {
			changed = true;
			// combine data
			factory.mergeNodeData(parent, child);
			// the single inEdge must come from parent, collapse child into parent
			final Edge parentChildEdge = child.getInEdges().get(0);
			g.removeEdge(parentChildEdge);
			// redirect child's out edges
			// clone, because we will modify the graph during traversal
			final ArrayList<Edge> childOutEdges = new ArrayList<Edge>(child.getOutEdges());
			for (final Edge childOutEdge : childOutEdges) {
				// remove
				g.removeEdge(childOutEdge);
				// only create a new edge if parent doesn't already contain it
				boolean parentAlreadyContainsThisOutEdge = false;
				for (final Edge parentOutEdge: parent.getOutEdges()) {
					if (parentOutEdge.getTarget()==childOutEdge.getTarget()) {
						parentAlreadyContainsThisOutEdge = true;
						break;
					}
				}
				if (!parentAlreadyContainsThisOutEdge) {
					final Edge newEdge = childOutEdge.clone();
					newEdge.setSource(parent);
					newEdge.setTarget(childOutEdge.getTarget());
					g.addEdge(newEdge);
				}
			}
			// redirect child's in edges
			// clone, because we will modify the graph during traversal
			final ArrayList<Edge> childInEdges = new ArrayList<Edge>(child.getInEdges());
			for (final Edge childInEdge : childInEdges) {
				// remove
				g.removeEdge(childInEdge);
				// only create a new edge if parent doesn't already contain it
				boolean parentAlreadyContainsThisInEdge = false;
				for (final Edge parentInEdge: parent.getInEdges()) {
					if (parentInEdge.getSource()==childInEdge.getSource()) {
						parentAlreadyContainsThisInEdge = true;
						break;
					}
				}
				if (!parentAlreadyContainsThisInEdge) {
					final Edge newEdge = childInEdge.clone();
					newEdge.setSource(childInEdge.getSource());
					newEdge.setTarget(parent);
					g.addEdge(newEdge);
				}
			}
			// remove child
			g.removeNode(child);
		}
		return changed;
	}
	
	private static boolean hasSingleNode(final FlowGraph g) {
		return g.getNodes().size()==1 && g.getNodes().get(0) == g.getEntryNode();
	}

}
