package ch.usi.inf.sape.graph.algorithms;

import java.util.HashMap;
import java.util.Set;

import ch.usi.inf.sape.graph.Edge;
import ch.usi.inf.sape.graph.EdgeFilter;
import ch.usi.inf.sape.graph.FlowGraph;
import ch.usi.inf.sape.graph.Node;
import ch.usi.inf.sape.graph.loop.Loop;
import ch.usi.inf.sape.graph.loop.LoopForest;


/**
 * Implementation of Ramlingam02 (Steensgaard93 variant) of a loop forest finder for potentially non-reducible flow graphs.
 * 
 * @author Matthias.Hauswirth@usi.ch
 */
public final class LoopForestFinder {

	public static boolean trace = false;
	
	
	//private static final class NodeProperties {
	//	public Loop innermostLoop;
	//}
	
	//private final HashMap<Node,NodeProperties> nodePropertiesByNode;
	private HashMap<Node,Loop> loopByNode;
	
	private EdgeFilter edgeFilter;
	
	
	public LoopForestFinder() {
	}
	
	public LoopForest find(final FlowGraph g) {
		//nodePropertiesByNode = new HashMap<Node, NodeProperties>();
		loopByNode = new HashMap<Node, Loop>();
		edgeFilter = new EdgeFilter();
		if (trace) System.out.println("LoopForestFinder.find("+g+")");		
		if (trace) System.out.println("Nodes: "+g.getNodes().size());
		if (trace) System.out.println("Edges: "+g.getEdges().size());
		final LoopForest f = new LoopForest();
		if (trace) System.out.println("Find roots (removed edges: "+edgeFilter.size()+")");
		final TarjanStronglyConnectedComponents sccFinder = new TarjanStronglyConnectedComponents();
		Set<StronglyConnectedComponent> sccs = sccFinder.getStronglyConnectedComponents(g, edgeFilter);
		if (trace) System.out.println("  "+sccs.size()+" SCCs found");
		for (final StronglyConnectedComponent scc : sccs) {
			if (isNonTrivial(scc)) {
				final Loop l = new Loop(null);
				f.addRoot(l);
				for (final Node node : scc.getNodes()) {
					l.addNode(node);
					setLoop(node, l);
				}
				addHeaders(l, g);
				removeLoopbackEdges(l);
			}
		}
		int level = 1;
		int previousSccCount = sccs.size();
		if (trace) System.out.println("Find level "+level+" (removed edges: "+edgeFilter.size()+")");
		// sccFinder has to IGNORE the removed edges!
		sccs = sccFinder.getStronglyConnectedComponents(g, edgeFilter);
		if (trace) System.out.println("  "+sccs.size()+" SCCs found");
		while (containsNonTrivial(sccs)) {
			
			if (previousSccCount==sccs.size()) {
				debug(g, sccs);
			}
			previousSccCount = sccs.size();
			for (final StronglyConnectedComponent scc : sccs) {
				if (isNonTrivial(scc)) {
					// currently, any node in scc will map to the parent loop
					final Loop parent = getLoop(scc.getNodes().get(0));
					final Loop l = new Loop(parent);
					for (final Node node : scc.getNodes()) {
						l.addNode(node);
						setLoop(node, l);
					}
					addHeaders(l, g);
					removeLoopbackEdges(l);
				}
			}
			level++;
			if (trace) System.out.println("Find level (removed edges: "+edgeFilter.size()+")");
			// sccFinder has to IGNORE the removed edges!
			sccs = sccFinder.getStronglyConnectedComponents(g, edgeFilter);
			if (trace) System.out.println("  "+sccs.size()+" SCCs found");
		}
		return f;
	}
	
	private void debug(FlowGraph g, Set<StronglyConnectedComponent> sccs) {
		int id = 0;
		for (final StronglyConnectedComponent scc : sccs) {
			if (isNonTrivial(scc)) {
				id++;
				System.out.println("Non-trivial SCC: "+id);
				for (Node n : scc.getNodes()) {
					System.out.println("  "+n);
					for (Edge e : n.getOutEdges()) {
						System.out.println("  out -> "+(edgeFilter.passes(e)?"":"(removed) ")+e.getTarget());
					}
					for (Edge e : n.getInEdges()) {
						System.out.println("  in  <- "+(edgeFilter.passes(e)?"":"(removed) ")+e.getTarget());
					}
				}
			}
		}
		System.exit(2);

	}
	
	private boolean isNonTrivial(final StronglyConnectedComponent scc) {
		assert scc.getNodes().size()>0;
		if (scc.getNodes().size()>1) {
			return true;
		} else {
			return hasSelfLoop(scc.getNodes().iterator().next());
		}
	}
	
	private boolean containsNonTrivial(final Set<StronglyConnectedComponent> sccs) {
		for (final StronglyConnectedComponent scc : sccs) {
			if (isNonTrivial(scc)) {
				return true;
			}
		}
		return false;
	}
	
	private boolean hasSelfLoop(final Node n) {
		for (final Edge e : edgeFilter.filter(n.getOutEdges())) {
			if (e.getTarget()==n) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Loop already has all body nodes set.
	 * Now identify its header nodes.
	 * 
	 * @param l
	 */
	private void addHeaders(final Loop l, final FlowGraph g) {
		//System.out.println("addHeaders");
		// use Steensgaard's approach (headers == entry nodes)
		for (final Node candidate : l.getBody()) {
			//System.out.println(candidate);
			if (candidate==g.getEntryNode()) {
				l.addHeader(candidate);
				//System.out.println(" - header (graph entry)");
			} else {
				for (final Edge e : edgeFilter.filter(candidate.getInEdges())) {
					// is source outside loop?
					//if (!l.getBody().contains(e.getSource())) {
					if (l!=getLoop(e.getSource())) {
						if (!l.getHeaders().contains(candidate)) {
							l.addHeader(candidate);
						}
						//System.out.println(" - header");
					}
				}
			}
		}
	}
	
	private void removeLoopbackEdges(final Loop l) {
		for (final Node header : l.getHeaders()) {
			for (final Edge e : edgeFilter.filter(header.getInEdges())) {
				// is source inside loop?
				//if (l.getBody().contains(e.getSource())) {
				if (l==getLoop(e.getSource())) {
					// mark edge as removed, but don't remove
					// (because we need to keep the graph in tact)
					edgeFilter.addRemovedEdge(e);
				}
			}
		}
	}
	
	
	//--- manage node properties
	private void setLoop(final Node node, final Loop loop) {
		//final NodeProperties p = new NodeProperties();
		//p.innermostLoop = loop;
		//nodePropertiesByNode.put(node, p);
		loopByNode.put(node, loop);
	}
	
	public Loop getLoop(final Node node) {
		return loopByNode.get(node);
		//final NodeProperties p = nodePropertiesByNode.get(node);
		//if (p!=null) {
		//	return p.innermostLoop;
		//} else {
		//	return null;
		//}
	}
	
}
