package ch.usi.inf.sape.graph.algorithms;

import java.util.Set;

import ch.usi.inf.sape.graph.Edge;
import ch.usi.inf.sape.graph.EdgeFilter;
import ch.usi.inf.sape.graph.FlowGraph;
import ch.usi.inf.sape.graph.Node;
import ch.usi.inf.sape.graph.loop.Loop2;
import ch.usi.inf.sape.graph.loop.LoopForest2;


/**
 * Implementation of Ramlingam02 (Steensgaard93 variant) of a loop forest finder for potentially non-reducible flow graphs.
 * 
 * SECOND VERSION OF IMPLEMENTATION, USING DIFFERENT LOOP FOREST REPRESENTATION!
 * 
 * @author Matthias.Hauswirth@usi.ch
 */
public final class LoopForestFinder2 {

	public static boolean trace = false;
	
	private EdgeFilter edgeFilter;
	
	
	public LoopForestFinder2() {
	}
	
	public LoopForest2 find(final FlowGraph g) {
		edgeFilter = new EdgeFilter();
		final LoopForest2 forest = new LoopForest2(g, edgeFilter);
		if (trace) System.out.println("LoopForestFinder.find("+g+")");		
		if (trace) System.out.println("Nodes: "+g.getNodes().size());
		if (trace) System.out.println("Edges: "+g.getEdges().size());
		int level = 0;
		int previousSccCount = -1;
		final TarjanStronglyConnectedComponents sccFinder = new TarjanStronglyConnectedComponents();
		if (trace) System.out.println("Find level "+level+" (removed edges: "+edgeFilter.size()+")");
		// sccFinder has to IGNORE the removed edges!
		Set<StronglyConnectedComponent> sccs = sccFinder.getStronglyConnectedComponents(g, edgeFilter);
		if (trace) System.out.println("  "+sccs.size()+" SCCs found");
		while (containsNonTrivial(sccs)) {
			if (previousSccCount==sccs.size()) {
				debug(g, sccs);
			}
			previousSccCount = sccs.size();
			for (final StronglyConnectedComponent scc : sccs) {
				if (isNonTrivial(scc)) {
					final Loop2 loop = forest.addLoop(scc);
					removeLoopbackEdges(loop, forest);
				}
			}
			level++;
			if (trace) System.out.println("Find level "+level+" (removed edges: "+edgeFilter.size()+")");
			// sccFinder has to IGNORE the removed edges!
			sccs = sccFinder.getStronglyConnectedComponents(g, edgeFilter);
			if (trace) System.out.println("  "+sccs.size()+" SCCs found");
		}
		return forest;
	}
	
	private void debug(FlowGraph g, Set<StronglyConnectedComponent> sccs) {
		int id = 0;
		for (final StronglyConnectedComponent scc : sccs) {
			if (isNonTrivial(scc)) {
				id++;
				System.out.println("Non-trivial SCC: "+id);
				for (Node n : scc.getNodes()) {
					System.out.println("  "+n);
					for (Edge e : n.getOutEdges()) {
						System.out.println("  out -> "+(edgeFilter.passes(e)?"":"(removed) ")+e.getTarget());
					}
					for (Edge e : n.getInEdges()) {
						System.out.println("  in  <- "+(edgeFilter.passes(e)?"":"(removed) ")+e.getTarget());
					}
				}
			}
		}
		System.exit(2);

	}
	
	private boolean isNonTrivial(final StronglyConnectedComponent scc) {
		assert scc.getNodes().size()>0;
		if (scc.getNodes().size()>1) {
			return true;
		} else {
			return hasSelfLoop(scc.getNodes().iterator().next());
		}
	}
	
	private boolean containsNonTrivial(final Set<StronglyConnectedComponent> sccs) {
		for (final StronglyConnectedComponent scc : sccs) {
			if (isNonTrivial(scc)) {
				return true;
			}
		}
		return false;
	}
	
	private boolean hasSelfLoop(final Node n) {
		for (final Edge e : edgeFilter.filter(n.getOutEdges())) {
			if (e.getTarget()==n) {
				return true;
			}
		}
		return false;
	}
	
	
	private void removeLoopbackEdges(final Loop2 l, final LoopForest2 forest) {
		for (final Node header : l.getHeaders()) {
			for (final Edge e : edgeFilter.filter(header.getInEdges())) {
				// is source inside loop?
				if (l==forest.getLoop(e.getSource())) {
					// mark edge as removed, but don't remove
					// (because we need to keep the graph in tact)
					edgeFilter.addRemovedEdge(e);
				}
			}
		}
	}
	
}
