package ch.usi.inf.sape.graph.algorithms;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import ch.usi.inf.sape.graph.Edge;
import ch.usi.inf.sape.graph.EdgeFilter;
import ch.usi.inf.sape.graph.FlowGraph;
import ch.usi.inf.sape.graph.Node;


/**
 * Determine the set of strongly connected components in the given graph.
 * 
 * @author Matthias.Hauswirth@usi.ch
 */
public final class TarjanStronglyConnectedComponents {

	private static final class NodeProperties {
		public int index = -1; // DFS numbering
		public int lowLink = -1;
		public boolean onStack = false; // currently stored on stack
	}
	
	private HashMap<Node, NodeProperties> properties;
	private HashSet<StronglyConnectedComponent> result;
	
	private int index;
	private ArrayList<Node> stack;
	
	
	public TarjanStronglyConnectedComponents() {
	}
	
	public Set<StronglyConnectedComponent> getStronglyConnectedComponents(final FlowGraph g) {
		return getStronglyConnectedComponents(g, new EdgeFilter());
	}
	
	/**
	 * Compute the strongly connected components in the given FlowGraph using Tarjan's algorithm.
	 * 
	 * @param g The FlowGraph for which we want to find SCCs.
	 * @return
	 */
	public Set<StronglyConnectedComponent> getStronglyConnectedComponents(final FlowGraph g, final EdgeFilter edgeFilter) {
		result = new HashSet<StronglyConnectedComponent>();
		initializeProperties(g);
		index = 0;
		stack = new ArrayList<Node>();
		for (final Node n : g.getNodes()) {
			if (getIndex(n)==-1) {
				traverseNode(n, edgeFilter);
			}
		}
		return result;
	}
	
	private void traverseNode(final Node n, final EdgeFilter edgeFilter) {
		setIndex(n, index);
		setLowLink(n, index);
		index++;
		push(n);
		for (final Edge e : edgeFilter.filter(n.getOutEdges())) {
			final Node n1 = e.getTarget();
			if (getIndex(n1)==-1) {
				traverseNode(n1, edgeFilter);
				setLowLink(n, Math.min(getLowLink(n), getLowLink(n1)));
			} else if (isOnStack(n1)) {
				setLowLink(n, Math.min(getLowLink(n), getIndex(n1)));
			}
		}
		if (getLowLink(n)==getIndex(n)) {
			//System.out.println("SCC:");
			final StronglyConnectedComponent scc = new StronglyConnectedComponent();
			result.add(scc);
			Node p;
			do {
				p = pop();
				//System.out.println(p);
				scc.addNode(p);
			} while (p!=n);
		}
	}
	
	
	//--- manage node properties
	private int getIndex(final Node n) {
		return properties.get(n).index;
	}
	
	private void setIndex(final Node n, final int index) {
		properties.get(n).index = index;
	}
	
	private int getLowLink(final Node n) {
		return properties.get(n).lowLink;
	}
	
	private void setLowLink(final Node n, final int lowLink) {
		properties.get(n).lowLink = lowLink;
	}
	
	private boolean isOnStack(final Node n) {
		return properties.get(n).onStack;
	}
	
	private void push(final Node n) {
		stack.add(n);
		properties.get(n).onStack = true;
	}
	
	private Node pop() {
		final Node n = stack.remove(stack.size()-1);
		properties.get(n).onStack = false;
		return n;
	}

	private void initializeProperties(final FlowGraph g) {
		properties = new HashMap<Node, NodeProperties>(g.getNodes().size());
		for (final Node n : g.getNodes()) {
			properties.put(n, new NodeProperties());
		}
	}
	
}
