package ch.usi.inf.sape.graph.dot;

import java.util.List;

import ch.usi.inf.sape.graph.Edge;
import ch.usi.inf.sape.graph.FlowGraph;
import ch.usi.inf.sape.graph.Node;
import ch.usi.inf.sape.graph.loop.Loop;
import ch.usi.inf.sape.graph.loop.Loop2;
import ch.usi.inf.sape.graph.loop.LoopForest;
import ch.usi.inf.sape.graph.loop.LoopForest2;


/**
 * Write a given FlowGraph or LoopForest to a dot file.
 *  
 * @author Matthias.Hauswirth@usi.ch
 */
public class DotWriter {

	public static String renderToDot(final FlowGraph g) {
		return renderToDot(g, new DefaultNodeAttributeStringProvider(), new DefaultEdgeAttributeStringProvider());
	}

	public static String renderToDot(final FlowGraph g, final NodeAttributeStringProvider nodeAttributeStringProvider, final EdgeAttributeStringProvider edgeAttributeStringProvider) {
		final StringBuffer sb = new StringBuffer("digraph ");
		sb.append('\"');
		sb.append(g.toString());
		sb.append('\"');
		sb.append(" {\n");
		for (final Node node : g.getNodes()) {
			sb.append("  ");
			sb.append('\"');
			sb.append(node.hashCode());
			sb.append('\"');
			sb.append(nodeAttributeStringProvider.getAttributeString(g, node));
			sb.append("\n");
		}
		for (final Edge edge : g.getEdges()) {
			sb.append("  ");
			sb.append('\"');
			sb.append(edge.getSource().hashCode());
			sb.append('\"');
			sb.append(" -> ");
			sb.append('\"');
			sb.append(edge.getTarget().hashCode());
			sb.append('\"');
			sb.append(edgeAttributeStringProvider.getAttributeString(g, edge));
			sb.append("\n");
		}
		sb.append("}\n");
		return sb.toString();
	}
	
	
	//----------------
	
	public static String renderLoopForest(final FlowGraph g, final LoopForest f) {
		final StringBuffer sb = new StringBuffer("digraph ");
		sb.append('\"');
		sb.append(g.toString());
		sb.append(" Loop Forest\"");
		sb.append(" {\n");
		for (final Loop l : f.getTreeRoots()) {
			renderLoopTree(l, sb);
		}
		sb.append("}\n");
		return sb.toString();
	}
	
	private static void renderLoopTree(final Loop l, final StringBuffer sb) {
		sb.append(" ");
		sb.append(getLoopId(l));
		sb.append(" [label=\"head: ");
		for (final Node h : l.getHeaders()) {
			sb.append(h+" ");
		}
		sb.append("\\nbody: ");
		for (final Node b : l.getBody()) {
			sb.append(b+" ");
		}
		sb.append("\"]\n");
		for (final Loop c : l.getChildren()) {
			sb.append(" ");
			sb.append(getLoopId(l));
			sb.append(" -> ");
			sb.append(getLoopId(c));
			sb.append("\n");
			renderLoopTree(c, sb);
		}
	}
	
	private static String getLoopId(final Loop l) {
		StringBuffer sb = new StringBuffer("\"");
		for (final Node h : l.getHeaders()) {
			sb.append(h+" ");
		}
		sb.append("\"");
		return sb.toString();
	}

	
	

	public static String renderLoopForest2(final FlowGraph g, final LoopForest2 f, final NodeAttributeStringProvider nodeAttributeStringProvider) {
		final StringBuffer sb = new StringBuffer("digraph ");
		sb.append('\"');
		sb.append(g.toString());
		sb.append(" Loop Forest2\"");
		sb.append(" {\n");
		for (final Node node : g.getNodes()) {
			sb.append("  \""+node.toString()+"\" "+nodeAttributeStringProvider.getAttributeString(g, node)+"\n");
			final Loop2 loop = f.getLoop(node);
			sb.append("  \""+loop.getId()+"\" -> \""+node.toString()+"\"\n");
		}
		sb.append("\n");
		renderLoop2(sb, f.getRoot(), f);
		sb.append("}\n");
		return sb.toString();
	}
	
	private static void renderLoop2(final StringBuffer sb, final Loop2 loop, final LoopForest2 f) {
		sb.append("  \""+loop.getId()+"\"");
		if (loop==f.getRoot()) {
			sb.append(" [label=\"FOREST\",shape=ellipse]\n");
		} else {
			sb.append(" [label=\"");
			sb.append(getLoop2Label(loop));
			sb.append("\",shape=rectangle]\n");
		}
		for (final Loop2 child : loop.getChildren()) {
			renderLoop2(sb, child, f);
			sb.append("  \""+loop.getId()+"\" -> \""+child.getId()+"\"\n");
		}
	}
	
	private static String getLoop2Label(final Loop2 l) {
		final StringBuffer sb = new StringBuffer("headers: \\n");
		final List<Node> headers = l.getHeaders();
		for (int i=0; i<headers.size(); i++) {
			final Node header = headers.get(i);
			sb.append(header.toString());
			if (i<headers.size()-1) {
				sb.append("\\n");
			}
		}
		return sb.toString();
	}
	
}
