package ch.usi.inf.sape.graph.loop;

import java.util.HashMap;
import java.util.List;

import ch.usi.inf.sape.graph.Edge;
import ch.usi.inf.sape.graph.EdgeFilter;
import ch.usi.inf.sape.graph.FlowGraph;
import ch.usi.inf.sape.graph.Node;
import ch.usi.inf.sape.graph.algorithms.StronglyConnectedComponent;


/**
 * The new, more memory-efficient representation of a loop forest.
 * 
 * Use this instead of LoopForest.
 * 
 * This class works in concert with the LoopForestFinder2 class.
 * This class holds the result, but it is also participating in construction of the forest.
 * 
 * @author Matthias.Hauswirth@usi.ch
 */
public class LoopForest2 {

	private final FlowGraph graph;
	private final EdgeFilter edgeFilter;
	private HashMap<Node,Loop2> loopByNode;
	private Loop2 root;
	
	
	
	/**
	 * Invoked by LoopForestFinder2. Don't use otherwise.
	 */
	public LoopForest2(final FlowGraph graph, final EdgeFilter edgeFilter) {
		this.graph = graph;
		this.edgeFilter = edgeFilter;
		root = new Loop2(null);
		final List<Node> nodes = graph.getNodes();
		loopByNode = new HashMap<Node, Loop2>(nodes.size());
		// initially, all nodes point to the root loop
		for (final Node node : nodes) {
			loopByNode.put(node, root);
		}
	}
	
	/**
	 * Root is not (!) a loop but represents the entire forest!
	 * @return the "loop" representing the entire forest (its children are the roots of individual loop nest trees)
	 */
	public Loop2 getRoot() {
		return root;
	}
	
	public Loop2 addLoop(final StronglyConnectedComponent scc) {
		// at this point, any node in scc will map to the parent loop
		final Loop2 parent = getLoop(scc.getNodes().get(0));
		final Loop2 loop = new Loop2(parent);
		// now, make all nodes in the scc point to this new loop
		for (final Node node : scc.getNodes()) {
			loopByNode.put(node, loop);
		}
		// during construction, loops don't store any body nodes, only header nodes
		addHeaders(loop, scc);
		return loop;
	}
	
	/**
	 * Identify and add header nodes to a loop (using Steensgaard's approach).
	 * 
	 * @param loop The loop to add headers to
	 * @param scc The SCC containing the loop's body nodes
	 */
	private void addHeaders(final Loop2 loop, final StronglyConnectedComponent scc) {
		//System.out.println("addHeaders");
		final List<Node> body = scc.getNodes();
		// use Steensgaard's approach (headers == entry nodes)
		for (final Node candidate : body) {
			//System.out.println(candidate);
			if (candidate==graph.getEntryNode()) {
				loop.addHeader(candidate);
				//System.out.println(" - header (graph entry)");
			} else {
				for (final Edge e : edgeFilter.filter(candidate.getInEdges())) {
					// is source outside loop?
					//if (!l.getBody().contains(e.getSource())) {
					if (loop!=getLoop(e.getSource())) {
						if (!loop.getHeaders().contains(candidate)) {
							loop.addHeader(candidate);
						}
						//System.out.println(" - header");
					}
				}
			}
		}
	}
	
	/**
	 * Given a Node from the FlowGraph on which this LoopForest2 is built,
	 * return the Loop2 that contains that Node.
	 * If the Node is not contained in any real loop,
	 * this returns the root (the artificial Loop2 that represents the entire forest).
	 */
	public Loop2 getLoop(final Node node) {
		return loopByNode.get(node);
	}

	/**
	 * Is the given Node a header in some loop?
	 */
	public boolean isHeader(final Node node) {
		// Note: a node cannot be in the header of TWO loops.
		// If it is in the header of a loop, then it is the header of the innermost loop containing the node
		// (which is the loop we get with loopByNode).
		return loopByNode.get(node).getHeaders().contains(node);
	}
		
}
