package usi.instrumentation.tree;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;

import usi.instrumentation.CCT;
import usi.instrumentation.stack.Node;
import usi.instrumentation.stack.Tree;

public abstract class AbstractTree implements TreeIR {
	
	private BufferedWriter bw;
	private int id = 0;
	protected HashSet<String> repetitions;
	
    public void DFS (final CCT cur, final String name) {
		Tree<String> tree = cur.tree;
		dotPrinter(name + "-", id);
		id++;
		
		List<Node<String>> treeList = tree.toList();
		int vertexCount = treeList.size();
		HashMap<Node<String>, VertexState> state = new HashMap<Node<String>, VertexState>();
          for (int i = 0; i < vertexCount; i++)
                state.put(treeList.get(i), VertexState.White);
       runDFS(tree.getRootElement(), state);
       for (Node<String> node: tree.toList()) {
    	   drawNode(node);
    	   drawBackEdges(node);
    	   drawEdges(node);   
		}
       
       
       optimiseSubGraphs();
       drawSubGraphs();
       closePrinter();   

    }
    //remove undrawn nodes
    private void removeUndrawnSubGraphs() {
    	ArrayList<Integer> toRemove = new ArrayList<Integer>();
    	for (Subgraph sub: Subgraph.subgraphs.values()) {
    		for (Integer i: sub.getNodeIds()) {
    			if (!drawnNodes.contains(i)) {
    				toRemove.add(i);
    			}
    		}
    		if (toRemove.size()>0) {
    			sub.nodeIds.removeAll(toRemove);
    			toRemove.clear();
    		}
    	}
    }
    /*
     * this method re-groups clusters 
     * so that nodes that reside in multiple cluster
     * would end up only in the largest cluster
     */
    private void optimiseSubGraphs() {	
    	removeUndrawnSubGraphs();
    	ArrayList<Integer> toRemove = new ArrayList<Integer>();
    	ArrayList<Integer> toAdd = new ArrayList<Integer>();
    	for (Integer number : drawnNodes) {
    		if (Subgraph.subgraphs.containsKey(number)) {	
    			Subgraph sub = Subgraph.subgraphs.get(number);
    			for (Integer in: sub.getNodeIds()) {
    				if (in != number && !toRemove.contains(in)) {
    					if (Subgraph.subgraphs.containsKey(in)) {
    						Subgraph sub2 = Subgraph.subgraphs.get(in);
    						for (Integer out : sub2.getNodeIds()) {
    							if (!sub.getNodeIds().contains(out)) {
    								//sub.getNodeIds().add(out);
    								if (!toAdd.contains(out))
    									toAdd.add(out);
    							}
    						}
    						if (!toRemove.contains(in)) {
    							toRemove.add(in);
    						}
    					}
    				}
    			}
    			sub.getNodeIds().addAll(toAdd);
    			toAdd.clear();
    		}
    	}
    	/*for (Integer in: toRemove) {
    		Subgraph.subgraphs.remove(in);    		
    	}*/
    }
    
    private void drawSubGraphs() {
    
    	List<Integer> used = new ArrayList<Integer>();
    	List<Subgraph> subs = new ArrayList<Subgraph>();
    	for (Integer number : drawnNodes) {
    		if (Subgraph.subgraphs.containsKey(number)) {	
    			Subgraph sub = Subgraph.subgraphs.get(number);
    			subs.add(sub);
    			used.add(number);
    			for (Integer in: sub.getNodeIds()) {
    				used.add(in);
    			}
    		}
    	}
    	Object[] graphs = subs.toArray();
    	Arrays.sort(graphs);
    	for (int i=graphs.length; i>0; i--) {
    		Subgraph cur = (Subgraph)graphs[i-1];
    		try {
    			StringBuffer sb = new StringBuffer();
				sb.append("subgraph ");
				sb.append(cur.getName() + " {\n"); 
				sb.append("label=\""+ cur.getDesc() +"\", style=filled, color=\"lightgrey\"]\n");
				for (Integer id: cur.getNodeIds()) {
					
					if (used.contains(id)) {
						sb.append(id + ",\n");
						used.remove(id);
					} else {
						//skip drawing
						continue;
					}
				}
				sb.append("}\n");
				bw.write(sb.toString());
				
			} catch (IOException e) {
				
				e.printStackTrace();
			}
    	}
    }

    private void closePrinter() {
    	try {
    		bw.write("}");
			bw.flush();
			bw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
    	
    }
    
    private List<String> duplicates = new ArrayList<String>();

    public void runDFS (Node<String> node, HashMap<Node<String>, VertexState> state) {
          state.put(node, VertexState.Gray);
          if (node.getNumberOfChildren()!=0) {
        	  int updatedChildrenNumber = node.getNumberOfChildren();
        	  List<Node<String>> ch = new ArrayList<Node<String>>();
          	  for (int i = 0; i<updatedChildrenNumber; i++) {
          		  ch.add(node.children.get(i));
          	  }
          	  for (Node<String> child: ch) {  
      			  if (state.get(child) == VertexState.White) {
      				  runDFS(child, state);
      				  
      			  }  
      		  }
  
          }
          //if it is black and method we may eliminate it and shift all the children up
          state.put(node, VertexState.Black);
          //time to print node and where it points to
          if (putUp(node)) {
        	  if (node.data.contains("<")) {
        		  node.data = node.data.replace("<", "(").replace(">", ")");
        	  }
          }
    }
    
    protected abstract void drawBackEdges(Node<String> node);
    
    
    /*
     * returns true if it don't have to shift parents up
     * false otherwise
     */
    protected boolean putUp (Node<String> node) {
    	if (!isEligible(node) && node.father!=null) {
    		if (node.getNumberOfChildren()>0) {
    			boolean alreadyExists = false;
    			List<Node<String>> toRemove = new ArrayList<Node<String>>();
    			for (Node<String> child: node.children) {
    				for (Node<String> n: node.father.getChildren()) {
    					if (n.data.equals(child.data)) {
    						alreadyExists = true;
    						break;
    					}
    				}
    				if (!alreadyExists) {
    					node.father.addChild(child);
    					child.father = node.father;
    				} else {
    					toRemove.add(child);
    					alreadyExists = false;
    					if (child.getNumberOfChildren()>0) {
    						//add them to n if they are different
    						//potential bug to fix!
    					}
    				}
    			}
    			if (!toRemove.isEmpty()) {
    				
    				node.children.removeAll(toRemove);
    				toRemove.clear();
    			}
    		}
    		node.father.children.remove(node);
    		return false;
    	} 
    	return true;
    }
    
   abstract protected boolean isEligible (Node<String> node);
    
   protected void draw (int hash1, int hash2, String label) {
	   try {
			  bw.write(hash1 + "->" + hash2 + 
					  "[ label =\"" +  label + "\"]\n");
			  
		  } catch (IOException e) {
			  e.printStackTrace();
		  } 
   }
    
    private void drawEdges (Node<String> node) {
    	
    	if (node.getNumberOfChildren()!=0) {
    		for (int i = 0; i<node.getNumberOfChildren(); i++) {
      		  Node<String> child = node.children.get(i);
      		  if (isEligible(child)) {
      			  String str = "";
      			 /* if (child.extra.numberOfEntrancies!=0) {
      				  str = str + child.extra.numberOfEntrancies;
      			  }*/
      			if (!repetitions.contains(""+child.hashCode()+ "x"+node.hashCode())) {
					repetitions.add("" + child.hashCode() + "x" + node.hashCode());
					draw(node.hashCode(), child.hashCode(), str);       			
      			}
      		  }
    		}    		
    	}
    	
    }
    
    private List<Integer> drawnNodes = new ArrayList<Integer>();
        
    private void drawNode (Node<String> node) {
    	if (!drawnNodes.contains(node.hashCode())) {
    		drawnNodes.add(node.hashCode());
    	}
    	try {
    		String colour = "red";
    		   
    		if (node.data.contains("-")) {
    			colour = "grey";
    		} 
    		String text;
    		if (node.data.equals("ENTRY")) {
				colour = "white";
				text = node.data;
			
    		} else {
    			if (node.data.contains("/")) {
    				text = returnFormattedNodeName(node.data);    				
    			} else {
    				text = node.data;
    			}
    		}
    		
    		
    		
			bw.write(node.hashCode() + "[shape=record, style=filled, fillcolor="+ colour + ", label = \"" + 
					text + "\"]\n");
		} catch (IOException e) {
			e.printStackTrace();
		}
		
    }
    
    /**
     * when node is a method we should first factor out signature
     * @param name of the node
     * @return adjusted for dot node name
     */
    
    private String returnFormattedNodeName (String line) {
    	String text = "";
    	
    	if (line.contains("(")) {
    		line = line.substring(0, line.indexOf("("));
    	}
    	String head = line.substring(0, line.lastIndexOf("/"));
		String tail = line.substring(line.lastIndexOf("/"), line.length());
		String method = head.substring(head.lastIndexOf("/"), head.length()) + tail;
		
		String clazz = head.substring(0, head.lastIndexOf("/"));
		if (clazz.contains("/")) {
			text = clazz.substring(clazz.lastIndexOf("/"), clazz.length()) + "\\n"  + method;
		} else {
			text = clazz + "\\n"  + method;
		}
    	
    	return text;
    }
    
    public void dotPrinter(String name, long id) {
    	try {
			bw = new BufferedWriter(new FileWriter(new File(name + 
					+ id + ".dot")));
			bw.write("digraph G {\n");
		} catch (IOException e) {
			System.out.println("Nests-loop tree dot file failed");
			e.printStackTrace();
		}
    }
    
    public void drawData (int node, String data) {
    	
    	//java_lang_Object [shape=ellipse,label="java.lang\nObject", style=dotted]
    	
    	try {
    	
    		int hash = node ^ data.hashCode();
			bw.write(hash + "[shape=ellipse, label = \"" + 
					data + "\"" +  ", style=dotted]\n");
			draw(node, hash, "");
		} catch (IOException e) {
			e.printStackTrace();
		}
    }

}


